import argparse
import os
import time

import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.galil_axis.drivers import GalilAxis


def read_config(configfile):
    """ read and parse YAML config for axes.
    config file is str"""
    try:
        with open(configfile, 'r') as f:
            cfg = yaml.safe_load(f) or {}
    except Exception as e:
        raise RuntimeError(f"Failed to read YAML '{configfile}': {e}")

    gal = cfg.get('galil')
    if not isinstance(gal, dict):
        raise ValueError("Config missing top-level 'galil' section.")

    brakes = gal.get('brakes')
    if not isinstance(brakes, dict) or not brakes:
        raise ValueError("galil.brakes must be a non-empty dict of output map per axis.")

    motorsettings = gal.get('motorsettings')
    if not isinstance(motorsettings, dict) or not motorsettings:
        raise ValueError("galil.motorsettings must be a non-empty dict of mm/deg conversion values.")

    mparams = gal.get('motorconfigparams')
    if not isinstance(mparams, dict) or not mparams:
        raise ValueError("galil.motorconfigparams must be a non-empty dict of axes.")

    initparams = gal.get('initaxisparams')
    if not isinstance(initparams, dict) or not initparams:
        raise ValueError("galil.initaxisparams must be a non-empty dict of axes settings.")

    dwellparams = gal.get('dwell_times')
    if not isinstance(dwellparams, dict) or not dwellparams:
        raise ValueError("galil.dwell_times must be a non-empty dict of times for initializing axes.")

    return cfg


class GalilAxisAgent:
    """ Agent for controlling Galil axis motors used in SAT
    coupling optics instrument for passband measurements on-site.

    Args:
        ip (str): IP address for the Galil axis motor controller
        configfile (str): Path to .yaml config file containing
            axis and motor settings
        port (int, optional): TCP port for controller communication.
            Default is 23
    """

    def __init__(self, agent, ip, configfile, port=23):
        self.lock = TimeoutLock()
        self.agent = agent
        self.log = agent.log
        self.ip = ip
        self.port = port
        self.configfile = configfile

        self.initialized = False
        self.take_data = False

        self.stage = None

        # Configuration state
        self.cfg = None
        self.motorsettings = None
        self.limitpol = None
        self.disable_limits_map = None
        self.disable_limit_axes = None
        self.axis_map = None
        self.brakes = None
        self.axes = None
        self.counts_per_mm = None
        self.counts_per_deg = None
        self.first_dwell = None
        self.sec_dwell = None

        # Register data feeds
        agg_params = {
            'frame_length': 60,
        }

        self.agent.register_feed('stage_status',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=1)

    @ocs_agent.param('input_config', type=bool, default=False)
    @ocs_agent.param('auto_acquire', type=bool, default=False)
    def init(self, session, params=None):
        """init(input_config=False, auto_acquire=False)

        **Task** - Initalize connection to Galil axis controller.

        Parameters:
            input_config (bool, optional): If True, init task will also
                run `input_configfile` task as part of agent initialization.
                Defaults to False.
            auto_acquire(bool):  If True, start acquisition immediately after
                initialization. Defaults to False.

        """
        if self.initialized:
            return True, "Already initialized"

        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                self.log.warn(f"Could not start init because {self.lock.job}"
                              "is already running")
                return False, "Could not acquire lock."

            # Get axes from config file to establish connection to galil stage controller
            try:
                cfg = read_config(self.configfile)
            except Exception as e:
                return False, f'Config load failed: {e}. Could not start connection.'

            self.cfg = cfg
            gal = self.cfg['galil']
            self.axes = list(gal['motorconfigparams'].keys())
            self.counts_per_mm = gal['motorsettings']['countspermm']
            self.counts_per_deg = gal['motorsettings']['countsperdeg']
            self.brakes = gal['brakes']['output_map']
            self.first_dwell = gal['dwell_times']['first_ms']
            self.sec_dwell = gal['dwell_times']['second_ms']

            # Establish connection
            self.stage = GalilAxis(ip=self.ip, port=self.port)

            # Test connection
            try:
                self.stage.get_data(self.axes)
            except ConnectionError:
                self.log.error("Could not establish connection to galil axis motor controller")
                return False, "Galil Axis Controller agent initialization failed"

        self.initialized = True

        # load configfile if provided as init argument
        if params['input_config']:
            ok, msg = self.input_configfile(session, {'configfile': self.configfile})
            self.log.info(msg)
            if not ok:
                return False, f'Config load failed: {msg}'

        # start data acquistion if requested
        if params['auto_acquire']:
            resp = self.agent.start('acq', params={})
            self.log.info(f'Response from acq.start(): {resp[1]}')

        return True, "Galil Axis Controller agent initialized"

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params):
        """acq()

        **Process** - Starts acquisition of data from the Galil Stage Controller.

        Parameters:
            test_mode (bool, optional): Run the process loop only once.
                This is meant only for testing. Default is False.

        Notes:
            The data collected is stored in session data in the structure::

                >> response.session['data']
                {'fields':
                    {'A_position': -1129.0, 'A_velocity': 0.0, 'A_torque': 0.0, 'A_position_error': 7.0,
                     'B_position': -542.0, 'B_velocity': 0.0, 'B_torque': 0.0, 'B_position_error': 14.0,
                     'C_position': 810637.0, 'C_velocity': -14983.0, 'C_torque': -1.0501, 'C_position_error': -558.0,
                     'D_position': 810711.0, 'D_velocity': -15240.0, 'D_torque': -1.0421, 'D_position_error': -580.0},
                     'timestamp': 1764009732.0866072}

        """
        with self.lock.acquire_timeout(0, job='acq') as acquired:
            if not acquired:
                self.log.warn(f"Could not start acq because {self.lock.job} is already running")
                return False, "Could not acquire lock."

            last_release = time.time()

            self.take_data = True

            pm = Pacemaker(1 / 3, quantize=False)
            while self.take_data:
                pm.sleep()
                # Reliqinuish sampling lock occassionally
                if time.time() - last_release > 1:
                    last_release = time.time()
                    if not self.lock.release_and_acquire(timeout=10):
                        self.log.warn(f"Failed to re-acquire sampling lock, "
                                      f"currently held by {self.lock.job}.")
                        continue

                try:
                    data = self.stage.get_data(self.axes)
                    if session.degraded:
                        self.log.info("Connection re-established.")
                        session.degraded = False
                except ConnectionError:
                    self.log.error("Failed to get data from galil stage controller. Check network connection.")
                    session.degraded = True
                    time.sleep(1)
                    continue

                axis_fields = {}
                for axis, stats in data.items():
                    for key, value in stats.items():
                        axis_fields[f"{axis}_{key}"] = value

                session.data = {"fields": axis_fields,
                                "timestamp": time.time()}

                pub_data = {'timestamp': time.time(),
                            'block_name': 'axes',
                            'data': axis_fields}

                self.agent.publish_to_feed('stage_status', pub_data)

                if params['test_mode']:
                    break

        self.agent.feeds['stage_status'].flush_buffer()

        return True, 'Acquisition exited cleanly.'

    def _stop_acq(self, session, params):
        """
        Stops acq process.
        """
        if self.take_data:
            self.take_data = False
            return True, 'requested to stop taking data.'
        else:
            return False, 'acq is not currently running'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('distance', type=float)
    @ocs_agent.param('movetype', type=str, choices=['linear', 'angular', 'encoder'])
    def set_relative_position(self, session, params):
        """set_relative_position(axis, distance, movetype)

        **Task** - Set the relative position location for a specified axis.

        Parameters:
            axis (str): Axis to command position. Ex: 'A'
            distance (float): Relative distance value, in millimeter, degrees,
                or raw counts, depending on `movetype`.
            movetype (str): 'linear' (mm) or 'angular' (deg), or 'encoder' for raw
                encoder counts.
        Notes:
            The conversion from millimeters to encoder counts is defined in the
            configuration file under `galil.motorsettings.countspermm` or
            `galil.motorsettings.countsperdeg`

        """
        axis = params['axis']
        distance = params['distance']
        movetype = params['movetype']
        with self.lock.acquire_timeout(timeout=5, job='set_relative_position') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            # Determine conversion factors
            if movetype == 'encoder':
                # use raw encoder units
                counts_per_unit = None
            elif movetype == 'linear':
                # use specific conversion factors
                counts_per_unit = self.counts_per_mm
            elif movetype == 'angular':
                counts_per_unit = self.counts_per_deg

            self.stage.set_relative_position(axis, distance,
                                             counts_per_unit=counts_per_unit)

        return True, f"Commanded {axis} to set to relative position: {distance}."

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('position', type=float)
    @ocs_agent.param('movetype', type=str, choices=['linear', 'angular', 'encoder'])
    def set_absolute_position(self, session, params):
        """set_absolute_position(axis, position, movetype)

        **Task** - Set the absolute position location for a specified axis.

        Parameters:
            axis (str): Axis to command position. Example: 'A'
            position (float): Absolute position distance value in mm, deg,
                or raw encoder counts.
            movetype (str): 'linear' (mm) or 'angular' (deg), or 'encoder' for raw
                encoder counts.

        Notes:
            The conversion from millimeters to encoder counts is defined in the
            configuration file under `galil.motorsettings.countspermm` or
            `galil.motorsettings.countsperdeg`
        """
        axis = params['axis']
        position = params['position']
        movetype = params['movetype']
        with self.lock.acquire_timeout(timeout=5, job='set_absolute_position') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running")
                return False, "Could not acquire lock"

            # Determine conversion factors
            if movetype == 'encoder':
                # use raw encoder units
                counts_per_unit = None
            elif movetype == 'linear':
                # use specific conversion factors
                counts_per_unit = self.counts_per_mm
            elif movetype == 'angular':
                counts_per_unit = self.counts_per_deg

            # Set the PA value
            self.stage.set_absolute_position(axis, position,
                                             counts_per_unit=counts_per_unit)

        return True, f"Commanded {axis} to set to absolute position: {position}"

    @ocs_agent.param('axis', type=str)
    def get_brake_status(self, session, params):
        """get_brake_status(axis)

        **Task** - Get brake status for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A', 'B').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_brake_status') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running")
                return False, "Could not acquire lock"

            output_num = self.brakes[axis]
            state, status = self.stage.get_brake_status(axis=axis, output_num=output_num)

        return True, f'Brake status for {axis} is {state}, {status}'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('state', type=str, choices=['engage', 'release'])
    def set_axis_brake(self, session, params):
        """set_axis_brake(axis, state)

        **Task** - Engages or releases the brake for the specified axis.

        Parameters:
            axis (str): Axis to control brake for. Ex. 'A'
            state (str): Desired brake state: 'engage' or 'release'.
        """
        axis = params['axis']
        state = params['state'].lower()

        with self.lock.acquire_timeout(timeout=5, job='set_axis_brake') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            brake_outputnum = self.brakes[axis]

            if state == 'engage':
                self.stage.engage_brake(output_num=brake_outputnum)
            elif state == 'release':
                self.stage.release_brake(output_num=brake_outputnum)

        return True, f"Commanded the brake for axis {axis} to be set to {state}."

    @ocs_agent.param('axis', type=str)
    def get_thermistor_voltage(self, session, params):
        """get_thermistor_voltage(axis)

        **Task** - Returns a given motor's thermistor reading in volts.

        Parameters:
            axis (str): Axis/motor to query (e.g. 'A')
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_thermistor_voltage') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running")
                return False, "Could not acquire lock"

            outputnum = self.brakes[axis]
            volt = self.stage.get_thermistor_voltage(axis=axis, output_num=outputnum)

        return True, f'Thermistor reading for motor {axis} ({outputnum}) is {volt} V.'

    @ocs_agent.param('axis', type=str)
    def begin_axis_motion(self, session, params):
        """begin_axis_motion(axis)

        **Task** - Moves the specified axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'

        """
        axis = params['axis']
        with self.lock.acquire_timeout(timeout=5, job='begin_axis_motion') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            status, msg = self.stage.begin_motion(axis)
            self.log.info(f'Status: {status}, {msg}')

        return True, f'Commanded {axis} to move.'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('motortype', type=int)
    def set_motor_type(self, session, params):
        """set_motor_type(axis)

        **Task** - Sets motor type for each axis, depending on servo motor features

        Parameters:
            axis (str): Specified axis. Ex. 'A'

        """
        axis = params['axis']
        motortype = params['motortype']
        with self.lock.acquire_timeout(timeout=5, job='set_motor_type') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_motor_type(axis, motortype)

        return True, f'Commanded motor type for {axis} to be set to {motortype}'

    @ocs_agent.param('axis', type=str)
    def get_motor_type(self, session, params):
        """get_motor_type(axis)

        **Task** - Queries the motor type for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A', 'B').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_motor_type') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_motor_type(axis)

        return True, f"Motor type for axis {axis}: {resp}"

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('errtype', type=int)
    def set_off_on_error(self, session, params):
        """set_off_on_error(axis)

        **Task** - Configure the off-on-error behavior for a specified axis.

        This task enables or disables the Galil controller's off-on-error function,
        which determines whether the motor power is shut off when a position error
        occurs.

        Parameters:
            axis (str): Axis to configure. Ex. 'A'
            errtype(int):  Error handling mode, as defined by the controller's
                disable/enable settings.

        """
        axis = params['axis']
        errtype = params['errtype']
        with self.lock.acquire_timeout(timeout=5, job='set_off_on_error') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_off_on_error(axis, errtype)

        return True, f'Commanded the off-on-error for {axis} to be set to {errtype}.'

    @ocs_agent.param('axis', type=str)
    def get_off_on_error(self, session, params):
        """get_off_on_error(axis)

        **Task** - Query the Off-On-Error (OE) state for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A', 'B', 'E').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_off_on_error') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            raw_val, human_state = self.stage.get_off_on_error(axis)

        return True, f"OE for {axis}: {human_state} (raw={raw_val})"

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=int)
    def set_amp_gain(self, session, params):
        """set_amp_gain(axis, val)

        **Task** - Set amplifier gain for internal amplifier per axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (int): Amplifier gain value to apply, as defined by
                controller's internal gain settings

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='set_amp_gain') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_amp_gain(axis, val)

        return True, f'Commanded amp gain for {axis} to be set to {val}.'

    @ocs_agent.param('axis', type=str)
    def get_amp_gain(self, session, params):
        """get_amp_gain(axis)

        **Task** - Query the amplifier gain value for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A', 'B', 'E').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_amp_gain') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_amp_gain(axis)

        return True, f"Amplifier gain for {axis}: {resp}"

    @ocs_agent.param('axis', type=str)
    def get_amp_currentloop_gain(self, session, params):
        """get_amp_currentloop_gain(axis)

        **Task** - Query the amplifier current loop gain value for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_amp_currentloop_gain') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_amp_currentloop_gain(axis)

        return True, f"Amplifier current loop gain for {axis} is: {resp}."

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=int)
    def set_amp_currentloop_gain(self, session, params):
        """set_amp_currentloop_gain(axis, val)

        **Task** - Set amplifier current loop gain for a given axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (int): Current-loop gain value to apply.

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='set_amp_currentloop_gain') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_amp_currentloop_gain(axis, val)

        return True, f'Commanded amp current loop gain to be set to {val} for axis {axis}.'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=float)
    def set_torque_limit(self, session, params):
        """set_torque_limit(axis, val)

        **Task** - Set motor torque limit per axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (float): Torque limit value in volts. Defines max
            allowable motor current, as specifiied by controller's
            internal settings.

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='set_torque_limit') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_torque_limit(axis, val)

        return True, f'Commanded torque limit for {axis} to be set to {val} volts'

    @ocs_agent.param('axis', type=str)
    def get_torque_limit(self, session, params):
        """get_torque_limit(axis)

        **Task** - Query motor torque limit for specified axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A').
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_motor_state') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_torque_limit(axis)

        return True, f"Torque limit for axis {axis} is {resp}."

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('state', type=str, choices=['enable', 'disable'])
    def set_motor_state(self, session, params):
        """set_motor_state(axis, state)

        **Task** - Enable or disable axis motor, and query its state.

        Parameters:
            axis (str): Axis to set motor state for (e.g. 'A', 'B', 'E').
            state (str): Desired motor state — 'enable' or 'disable'.
        """
        axis = params['axis']
        state = params['state']

        with self.lock.acquire_timeout(timeout=5, job='set_motor_state') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            self.stage.set_motor_state(axis, state)

        return True, f"Commanded motor {axis} to be {state}d."

    @ocs_agent.param('axis', type=str)
    def get_motor_state(self, session, params):
        """get_motor_state(axis)

        **Task** - Query and interpret whether a motor is ON or OFF for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A').

        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_motor_state') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            state, human_state = self.stage.get_motor_state(axis)

        return True, f"Motor {axis} state: {human_state} (raw={state})"

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=float)
    def set_magnetic_cycle(self, session, params):
        """set_magnetic_cycle(axis, val)

        **Task** - Set magnetic cycle value for motors with sinusoidal amplifiers for specific axis.

        Defines the length of the motor's magnetic cycle in encoder units.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (float): Magnetic cycle value to apply.

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='set_magnetic_cycle') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_magnetic_cycle(axis, val)

        return True, f'Commanded magnetic cycle for {axis} to be set to {val}.'

    @ocs_agent.param('t_first', type=int)
    @ocs_agent.param('t_second', type=int)
    def set_dwell_times(self, session, params):
        """set_dwell_times(t_first, t_second)

        **Task** - Define dwell times for the initialization task.

        This task defines the time between driving each axis motor to 2 different locations.

        Parameters:
            t_first (int): timing in milliseconds for driviing the motor to
                the first location. Ex: 1500
            t_second (int): timing in milliseconds for driving the motor to
                the second location. Ex: 1000

        """
        first = params['t_first']
        second = params['t_second']
        with self.lock.acquire_timeout(timeout=5, job='set_dwell_times') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_dwell_times(t_first=first, t_second=second)

        return True, f'Commanded dwell times for axes to be set to {first} and {second}ms, respectively.'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=float)
    def initialize_axis(self, session, params):
        """initialize_axis(axis, val)

        **Task** - Initialize axes configured with sinusoidal amplifiers.

        During this procedure, each motor is driven to two magnetic
        positions to establish the commutation angle required for motion.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (float): Torque command voltage to be applied during initialization. Ex: 3.0

        Notes:
            To run this task, you must run the `enable_sin_commutation`,
            `set_magnetic_cycle`, and `set_dwell_times` tasks first.

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='initialize_axis') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.initialize_axis(axis, val)

        return True, f'Commanded {axis} to be initialized to {val} setting'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=float)
    def define_position(self, session, params):
        """define_position(axis, val)

        **Task** - Resets the position of an axis encoder to a specifically set value.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (float): The value to set the current motor position as specified by user.
                (i.e., 0.00 if homing)

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='define_position') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.define_position(axis, val)

        return True, f'Commanded axis {axis} position to be set to {val}.'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('speed', type=float)
    def set_jog_speed(self, session, params):
        """set_jog_speed(axis, speed)

        **Task** - Defines the jog speed for the specified axis.

        This task will set the speed to move the axis continuously when ready to begin motion,
        but does not begin motion.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            speed (float): The value of the speed in raw encoder units

        Notes:
            The speed value here is not defined in terms of millimeters or
            degrees. User will have to define in encoder counts
            (i.e., `4000` for 4000 counts/sec).

        """
        axis = params['axis']
        speed = params['speed']
        with self.lock.acquire_timeout(timeout=5, job='set_jog_speed') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_jog_speed(axis, speed)

        return True, f'Commanded {axis} jog speed to be set to {speed}.'

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('speed', type=float)
    def set_speed(self, session, params):
        """set_speed(axis, speed)

        **Task** - Sets the speed for a subsequent  motion/move-to command
        (such as relative or absolute position motion).

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            speed (float): Specified speed value in raw encoder units

        Notes:
            The speed value here is not defined in terms of millimeters or
            degrees. User will have to define in encoder counts
            (i.e., `4000` for 4000 counts/sec).

        """
        axis = params['axis']
        speed = params['speed']
        with self.lock.acquire_timeout(timeout=5, job='set_speed') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_speed(axis, speed)

        return True, f'Commanded axis {axis} speed to be set to {speed}.'

    @ocs_agent.param('axis', type=str)
    def get_speed(self, session, params):
        """get_speed(axis)

        **Task** - Return the set speed for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A')
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_speed') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_speed(axis)

        return True, f"Speed for axis {axis}: {resp}"

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('acc', type=float)
    def set_acceleration(self, session, params):
        """set_acceleration(axis, acc)

        **Task** - Sets the acceleration for a subsequent motion/move-to command.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            acc (float): Specified acceleration value in raw encoder units.

        Notes:
            The acceleration value here is not defined in terms of millimters or degrees. User will have to define in encoder counts that are powers of 2. (i.e., '4096' for 4096 counts/sec^2).
        """
        axis = params['axis']
        acc = params['acc']
        with self.lock.acquire_timeout(timeout=5, job='set_acceleration') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_acceleration(axis, acc)

        return True, f'Commanded axis {axis} acceleration to be set to {acc}.'

    @ocs_agent.param('axis', type=str)
    def get_acceleration(self, session, params):
        """get_acceleration(axis)

        **Task** - Query acceleration for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A')
        """

        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_acceleration') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_acceleration(axis)

        return True, f"Acceleration for axis {axis}: {resp}"

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('decel', type=float)
    def set_deceleration(self, session, params):
        """set_deceleration(axis, decel)

        **Task** - Sets the deceleration for a subsequent motion/move-to command.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            decel (float): Deceleration in encoder units (counts/s^2)

        Notes:
            User will have to define in encoder counts/s^2 in powers of 2, (e.g. 8192).
        """
        axis = params['axis']
        decel = params['decel']

        with self.lock.acquire_timeout(timeout=5, job='set_deceleration') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            self.stage.set_deceleration(axis, decel)

        return True, f'Command axis {axis} deceleration to be set to {decel}'

    @ocs_agent.param('axis', type=str)
    def get_deceleration(self, session, params):
        """get_deceleration(axis)

        **Task** - Returns the deceleration for a given axis.

        Parameters:
            axis (str): Axis to query (e.g. 'A')
        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_deceleration') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running.")
                return False, "Could not acquire lock."

            resp = self.stage.get_deceleration(axis)

        return True, f"Deceleration for axis {axis}: {resp}"

    @ocs_agent.param('axis', type=str)
    def enable_sin_commutation(self, session, params):
        """enable_sin_commutation(axis)

        **Task** -  Enable sinusoidal commutation for a specified brushless axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'

        """
        axis = params['axis']
        with self.lock.acquire_timeout(timeout=5, job='enable_sin_commutation') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.enable_sin_commutation(axis)

        return True, f"Commanded sin commutation for axis {axis} to be enabled."

    @ocs_agent.param('axis', type=str)
    @ocs_agent.param('val', type=int)
    def set_limitswitch_mode(self, session, params):
        """set_limitswitch_mode(axis, val)

        **Task** - Enable or disable limit switch for a specified axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'
            val (int): 0 = both enabled; 1 = forward disabled; 2 = reverse disabled; 3 = both disabled

        """
        axis = params['axis']
        val = params['val']
        with self.lock.acquire_timeout(timeout=5, job='set_limitswitch_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_limitswitch_mode(axis, val)

        return True, f'Commanded limit switch mode for {axis} to be {val}'

    @ocs_agent.param('axis', type=str)
    def get_limitswitch_mode(self, session, params):
        """get_limitswitch_mode(axis)

        **Task** - Get the limit switch enable/disable setting for a given axis.

        Parameters:
            axis (str): Specified axis. Ex. 'A'

        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_limitswitch_mode') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            state, status = self.stage.get_limitswitch_mode(axis)

        return True, f'Limit switch for axis {axis} is set to {state, status}'

    @ocs_agent.param('polarity', type=int)
    def set_limitswitch_polarity(self, session, params):
        """set_limitswitch_polarity(polarity)

        **Task** - Configure the limit switch polarity for all axes.

        Parameters:
            polarity (int): Limit switch polarity value.
                0 is active-low (limit triggers when input goes low),
                1 is active-high (limit triggers when input goes high).

        """
        pol = params['polarity']
        with self.lock.acquire_timeout(timeout=5, job='set_limitswitch_polarity') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_limitswitch_polarity(pol)

        return True, f"Commanded limit switch polarity to be set to {pol}."

    @ocs_agent.param('_')
    def get_limitswitch_polarity(self, session, params):
        """get_limitswitch_polarity()

        **Task** - Query limit switch polarity setting.

        """
        with self.lock.acquire_timeout(timeout=5, job='get_limitswitch_polarity') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because {self.lock.job} is already running")
                return False, "Could not acquire lock"

            state, status = self.stage.get_limitswitch_polarity()

        return True, f'Limit switch polarity is {status}'

    @ocs_agent.param('axis', type=str)
    def stop_axis_motion(self, session, params):
        """stop_axis_motion(axis)

        **Task** - Stop motion for specified axis.

        Parameters:
              axis (str): Specified axis. Ex. 'A'

        """
        axis = params['axis']
        with self.lock.acquire_timeout(timeout=5, job='stop_axis_motion') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.stop_motion(axis)

        return True, f'Commanded {axis} a to stop motion.'

    @ocs_agent.param('order', type=str)
    def set_gearing(self, session, params):
        """set_gearing(order)

        **Task** - Configure electronic gearing relationships between follower and leader axes.

        Parameters:
            order (str): Comma-separated axis assignment string defining the leader/follower mapping (e.g., ',,,,,E' assigns E as the leader for F axis).

        """
        order = params['order']
        with self.lock.acquire_timeout(timeout=5, job='set_gearing') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_gearing(order)

        return True, f'Commanded gearing for leader/folower axes to be set to {order}.'

    @ocs_agent.param('order', type=str)
    def set_gearing_ratio(self, session, params):
        """set_gearing_ratio(order)

        **Task** - Set electronic gearing ratio between leader and follower axes.

        A ratio of 1 means the follower axis moves at the same speed as its
        leader axis.

        Parameters:
            order (str): Comma-separated string of gearing ratios for each axis, similar in format as `set_gearing` task. (e.g., ',,,,,1')

        """
        order = params['order']
        with self.lock.acquire_timeout(timeout=5, job='set_gearing_ratio') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            self.stage.set_gearing_ratio(order)

        return True, f'Commanded gearing ratio for follower axes to be set to {order}.'

    @ocs_agent.param('axis', type=str)
    def get_gearing_ratio(self, session, params):
        """get_gearing_ratio(axis)

        **Task** - Query the current electronic gearing ratio for a specified axis

        Parameters:
            axis (str): Axis to query (e.g., 'A')

        """
        axis = params['axis']
        with self.lock.acquire_timeout(timeout=5, job='get_gearing_ratio') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            resp = self.stage.get_gearing_ratio(axis=axis)

        return True, f"Gearing ratio for axis {axis} is: {resp}."

    @ocs_agent.param('axis', type=str)
    def get_gearing_lead(self, session, params):
        """get_gearing_lead(axis)

        **Task** - Query the leader axis for a given follower axis.

        Parameters:
            axis (str): Axis to query (e.g., 'A')

        """
        axis = params['axis']

        with self.lock.acquire_timeout(timeout=5, job='get_gearing_lead') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            resp = self.stage.get_gearing_lead(axis=axis)

        return True, f"Gearing lead for axis {axis} is: {resp}"

    @ocs_agent.param('configfile', type=str, default=None)
    def input_configfile(self, session, params=None):
        """input_configfile(configfile=None)

        **Task** - Upload Galil Axis Controller configuration file to initialize device
        and axes on device

        Parameters:
            configfile (str, optional):
                name of .yaml config file. Defaults to the file set in the site config.

        """
        configfile = params['configfile']
        if configfile is None:
            configfile = self.configfile
        if configfile is None:
            raise ValueError("No configfile specified")
        configfile = os.path.join(os.environ['OCS_CONFIG_DIR'], configfile)

        with self.lock.acquire_timeout(timeout=5, job='input_configfile') as acquired:
            if not acquired:
                self.log.warn(f"Could not start Task because "
                              f"{self.lock.job} is already running")
                return False, "Could not acquire lock"

            cfg = read_config(configfile)
            try:
                gal = cfg['galil']
                self.cfg = cfg
                self.motorsettings = gal['motorsettings']
                self.limitpol = gal['limitsetting']['polarity']
                self.disable_limits_map = gal['limitsetting']['disable_limits']
                self.disable_limit_axes = list(self.disable_limits_map.keys())
                self.axis_map = gal['motorconfigparams']
                self.axes = list(self.axis_map.keys())
                self.brakes = gal['brakes']['output_map']
                self.first_dwell = gal['dwell_times']['first_ms']
                self.sec_dwell = gal['dwell_times']['second_ms']
            except Exception as e:
                return False, f'Config parse error: {e}'

            # add line for limit switch polarity setting
            self.stage.set_limitswitch_polarity(pol=self.limitpol)

            # disable limit switches for the follower axes
            for x in self.disable_limit_axes:
                self.log.info(f"Disabling limit switch for axis {x} with setting {self.disable_limits_map[x]}")
                self.stage.set_limitswitch_mode(axis=x, val=self.disable_limits_map[x])
                time.sleep(1)

            # release brakes
            for x in self.brakes:
                self.log.info(f"Releasing brake for axis {x}")
                brake_outputnum = self.brakes[x]
                self.stage.release_brake(output_num=brake_outputnum)
                time.sleep(1)

            # set initialization dwell times if initialization setting is true
            initstate = gal['initaxisparams']['BA']
            if initstate == 'True':
                # set dwell times before initializing
                self.log.info(f"Setting dwell timies for initializing axes: {self.first_dwell} and {self.sec_dwell}")
                self.stage.set_dwell_times(t_first=self.first_dwell, t_second=self.sec_dwell)

            # Configure motor settings
            for a in self.axes:
                # set motor type
                self.log.info(f"Motor type set for axis {a} to be {self.axis_map[a]['MT']}")
                self.stage.set_motor_type(axis=a, motortype=self.axis_map[a]['MT'])

                # set off on error
                self.log.info(f"Setting off on error for axis {a}")
                self.stage.set_off_on_error(axis=a, errtype=self.axis_map[a]['OE'])

                # set amp gain
                self.log.info(f"Setting amp gain for axis {a} to be {self.axis_map[a]['AG']}")
                self.stage.set_amp_gain(axis=a, val=self.axis_map[a]['AG'])

                # set torque limit
                self.log.info(f"Setting torque limit for axis {a} to be {self.axis_map[a]['TL']}")
                self.stage.set_torque_limit(axis=a, val=self.axis_map[a]['TL'])

                # set current loop gain
                self.log.info(f"Setting amp current loop gain for axis {a} to be {self.axis_map[a]['AU']}")
                self.stage.set_amp_currentloop_gain(axis=a, val=self.axis_map[a]['AU'])

                if initstate == 'True':
                    # enable sin commutation
                    self.log.info(f"Enabling sinusoidal commutation for axis {a}")
                    self.stage.enable_sin_commutation(axis=a)

                    # set magnetic cycle
                    self.log.info(f"Setting magnetic cycle for axis {a}")
                    self.stage.set_magnetic_cycle(axis=a, val=gal['initaxisparams']['BM'])

                    # initialize
                    self.log.info(f"Driving the motors to initialize axis {a} for movement")
                    self.stage.initialize_axis(axis=a, val=gal['initaxisparams']['BZ'])
                    time.sleep(((int(self.first_dwell) + int(self.sec_dwell)) / 1000) * 2)

        return True, "Commanded input_configfile task to set motor controller settings."


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automaticall build documenation based on this function.


    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip')
    pgroup.add_argument('--port', default=23)
    pgroup.add_argument('--configfile')
    pgroup.add_argument('--input_config', type=bool)
    pgroup.add_argument('--mode', choices=['init', 'acq'])

    return parser


def main(args=None):
    parser = make_parser()
    args = site_config.parse_args(agent_class='GalilAxisAgent',
                                  parser=parser,
                                  args=args)

    # Automatically acquire data if requested (default)
    init_params = False
    if args.mode == 'init':
        init_params = {'auto_acquire': False,
                       'input_config': args.input_config}
    elif args.mode == 'acq':
        init_params = {'auto_acquire': True,
                       'input_config': args.input_config}

    # Call launcher function (initiates connection to appropriate
    # WAMP hub and realm).

    agent, runner = ocs_agent.init_site_agent(args)

    # create agent instance and run log creation
    galilaxis_agent = GalilAxisAgent(agent, args.ip, args.configfile)
    agent.register_task('init', galilaxis_agent.init, startup=init_params)
    agent.register_task('set_relative_position', galilaxis_agent.set_relative_position)
    agent.register_task('set_absolute_position', galilaxis_agent.set_absolute_position)
    agent.register_task('get_brake_status', galilaxis_agent.get_brake_status)
    agent.register_task('set_axis_brake', galilaxis_agent.set_axis_brake)
    agent.register_task('get_thermistor_voltage', galilaxis_agent.get_thermistor_voltage)
    agent.register_task('begin_axis_motion', galilaxis_agent.begin_axis_motion)
    agent.register_task('stop_axis_motion', galilaxis_agent.stop_axis_motion)
    agent.register_task('set_motor_type', galilaxis_agent.set_motor_type)
    agent.register_task('get_motor_type', galilaxis_agent.get_motor_type)
    agent.register_task('set_off_on_error', galilaxis_agent.set_off_on_error)
    agent.register_task('get_off_on_error', galilaxis_agent.get_off_on_error)
    agent.register_task('set_amp_gain', galilaxis_agent.set_amp_gain)
    agent.register_task('get_amp_gain', galilaxis_agent.get_amp_gain)
    agent.register_task('set_torque_limit', galilaxis_agent.set_torque_limit)
    agent.register_task('get_torque_limit', galilaxis_agent.get_torque_limit)
    agent.register_task('set_amp_currentloop_gain', galilaxis_agent.set_amp_currentloop_gain)
    agent.register_task('get_amp_currentloop_gain', galilaxis_agent.get_amp_currentloop_gain)
    agent.register_task('set_motor_state', galilaxis_agent.set_motor_state)
    agent.register_task('get_motor_state', galilaxis_agent.get_motor_state)
    agent.register_task('set_magnetic_cycle', galilaxis_agent.set_magnetic_cycle)
    agent.register_task('set_dwell_times', galilaxis_agent.set_dwell_times)
    agent.register_task('initialize_axis', galilaxis_agent.initialize_axis)
    agent.register_task('define_position', galilaxis_agent.define_position)
    agent.register_task('set_jog_speed', galilaxis_agent.set_jog_speed)
    agent.register_task('set_speed', galilaxis_agent.set_speed)
    agent.register_task('get_speed', galilaxis_agent.get_speed)
    agent.register_task('set_acceleration', galilaxis_agent.set_acceleration)
    agent.register_task('get_acceleration', galilaxis_agent.get_acceleration)
    agent.register_task('set_deceleration', galilaxis_agent.set_deceleration)
    agent.register_task('get_deceleration', galilaxis_agent.get_deceleration)
    agent.register_task('enable_sin_commutation', galilaxis_agent.enable_sin_commutation)
    agent.register_task('set_limitswitch_mode', galilaxis_agent.set_limitswitch_mode)
    agent.register_task('get_limitswitch_mode', galilaxis_agent.get_limitswitch_mode)
    agent.register_task('set_limitswitch_polarity', galilaxis_agent.set_limitswitch_polarity)
    agent.register_task('get_limitswitch_polarity', galilaxis_agent.get_limitswitch_polarity)
    agent.register_task('set_gearing', galilaxis_agent.set_gearing)
    agent.register_task('set_gearing_ratio', galilaxis_agent.set_gearing_ratio)
    agent.register_task('get_gearing_ratio', galilaxis_agent.get_gearing_ratio)
    agent.register_task('get_gearing_lead', galilaxis_agent.get_gearing_lead)
    agent.register_task('input_configfile', galilaxis_agent.input_configfile)
    agent.register_process('acq', galilaxis_agent.acq, galilaxis_agent._stop_acq)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import math
import select
import time

from socs.tcp import TCPInterface


class GalilAxis(TCPInterface):
    def __init__(self, ip, port=23, timeout=10):
        """Interface class for connecting to GalilStageController for SO SAT Coupling Optics."""
        self.ip = ip
        self.port = port
        self.timeout = timeout

        super().__init__(self.ip, self.port, self.timeout)

    def _drain_prompt(self):
        """
        Check for and drain any remaining data from the TCP buffer.
        Return True or False depending on whether a ':' Galil prompt appears.

        """
        drained = b""

        # check if anything in buffer
        rlist, _, _ = select.select([self.comm], [], [], 0.2)
        if rlist:
            chunk = self.recv()
            if chunk:
                drained += chunk

                # --- if ':' in buffer, galil is ready to receive command, return True ---
                if drained.endswith(b"\r:") or b":" in drained:
                    return True
                else:
                    # means more drainage necessary
                    print(f"Drained {len(drained)} bytes, but no ':' prompt seen.")
                    return False
        else:
            # no data pending in buffer -- ready to send new command!
            return True

    def _is_ready(self, max_attempts=3, delay=0.05):
        """Try up to max_attempts to drain prompt until controller is ready."""
        for attempt in range(max_attempts):
            if self._drain_prompt():
                return True
            time.sleep(delay)

        raise Exception(
            "':' not seen to indicate Galil ready to receive command. "
            "Flushed out buffer 3 times — likely needs to be flushed more."
        )

    def galil_command(self, command=None, axis=None, value=None,
                      expect_response=False, retries=3):
        """
        Send a command to the Galil controller and optionally return its response.

        """

        # --- Build command string ---
        if axis is not None and value is not None:
            cmd = f"{command}{axis}={value}"
        elif axis is not None:
            cmd = f"{command}{axis}"
        elif value is not None:
            cmd = f"{command}={value}"
        else:
            cmd = f"{command}"

        msg = f"{cmd}\r".encode("ascii")

        # check if Galil is ready to receive commands
        self._is_ready()

        self.send(msg)

        if not expect_response:
            resp = ''
            return resp
        else:
            for attempt in range(retries):
                resp = self.recv().decode("ascii", errors="ignore").strip(":\r\n")

                if resp in ("?", "??", ""):
                    self._drain_prompt()
                    self.send(b"TC1\r")  # ask galil why '?'
                    tc_resp = self.recv().decode("ascii", errors="ignore").strip(":\r\n ")

                    if tc_resp.startswith("5"):
                        print("TC1=5 (Input buffer full) — clearing input buffer")
                        self.send(b"CI -1;\r")
                    else:
                        print(f"TC error response: {tc_resp}")

                    time.sleep(0.1)
                    self.send(msg)
                    continue

                else:
                    break

            return resp

    def get_relative_position(self, axis, movetype=None, counts_per_mm=None, counts_per_deg=None):
        """
        Query the relative position set for a specified axis. Converts counts to physical units if
        movetype is 'linear' or 'angular'.

        """
        units_map = {'linear': 'mm', 'angular': 'deg'}
        units = units_map.get(movetype, '')

        # --- Query the controller using robust galil_command() ---
        value = self.galil_command("MG _PR", axis, expect_response=True)
        try:
            value = float(value)
        except Exception as e:
            print(f'Exception occurred, returning nan: {e}')
            return math.nan, units

        # If no movetype, just return raw counts
        if movetype is None:
            return value

        # Optional conversion if provided
        if movetype == 'linear' and counts_per_mm:
            value /= counts_per_mm
            return value, 'mm'
        elif movetype == 'angular' and counts_per_deg:
            value /= counts_per_deg
            return value, 'deg'

    def get_data(self, axes=None):
        """
        Query position (TP), velocity (TV), torque (TT), and position error (TE)
        for all axes in one batch, then subset results for the requested axes.

        """

        data = {}

        # Order of axes is always A–H on Galil (up to 8 possible)
        all_axes = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H']

        tp = [float(x) for x in self.galil_command('TP', expect_response=True).split(',')]
        tv = [float(x) for x in self.galil_command('TV', expect_response=True).split(',')]
        tt = [float(x) for x in self.galil_command('TT', expect_response=True).split(',')]
        te = [float(x) for x in self.galil_command('TE', expect_response=True).split(',')]

        # Determine how many valid axes are reported (Galil reports starting from A)
        n_axes = max(len(tp), len(tv), len(tt), len(te))
        active_axes = all_axes[:n_axes]

        if axes is None:
            axes = active_axes

        # --- Build dict per axis ---
        for i, axis in enumerate(active_axes):
            if axis not in axes:
                continue  # skip axes user doesn't care about

            data[axis] = {
                'position': tp[i],
                'velocity': tv[i],
                'torque': tt[i],
                'position_error': te[i]
            }

        return data

    def is_running(self, axis):
        """
        Checks if the axis is moving.

        """
        cmd = f'MG _BG{axis}'
        resp = self.galil_command(cmd, expect_response=True)
        return resp

    def begin_motion(self, axis):
        """
        Begin motion for the specified axis using the BG command.

        """
        self.galil_command(command="BG", axis=axis)

        state = self.is_running(axis)
        state = str(int(float(state)))
        if state == '1':
            msg = f'Axis {axis} is in motion.'
            return True, msg
        elif state == '0':
            msg = f'Axis {axis} is not moving. Retry command if needed.'
            return False, msg

    def set_relative_position(self, axis, distance, counts_per_unit=None):
        """
        Move axis by a relative distance (in calibration units such as
        mm or deg or raw encoder counts). If `counts_per_unit` is None,
        assume raw counts.

        """
        counts = distance if counts_per_unit is None else round(distance * counts_per_unit, 3)
        return self.galil_command("PR", axis=axis, value=counts)

    def set_absolute_position(self, axis, position, counts_per_unit=None):
        """
        Set absolute position for an axis in units or encoder counts.

        """
        counts = position if counts_per_unit is None else round(position * counts_per_unit, 3)
        return self.galil_command("PA", axis=axis, value=counts)

    def release_brake(self, output_num):
        """
        Release brake to axis by using the GalilDMC SB command which sets the digital
        output to 1.

        """
        resp = self.galil_command(command=f"SB {output_num}")
        return resp

    def engage_brake(self, output_num):
        """
        Engage the brake for the specified axis using the Galil CB command.

        """
        resp = self.galil_command(command=f"CB {output_num}")
        return resp

    def get_brake_status(self, axis, output_num):
        """
        Return brake status for axis via @OUT[n].

        """
        val = self.galil_command(command=f"MG @OUT[{output_num}]",
                                 expect_response=True)

        state = int(round(float(val)))

        if state == 1:
            status = "Brake Released"
        if state == 0:
            status = "Brake Engaged"

        return state, status

    def get_thermistor_voltage(self, axis, output_num):
        """
        Return thermistor voltage value for an axis via @AN[n]
        """
        val = self.galil_command(command=f"MG @AN[{output_num}]", expect_response=True)

        temp_voltage = float(val)

        return temp_voltage

    def get_motor_type(self, axis):
        """
        Return motor type for given axis via MG _MT{axis}.

        """
        resp = self.galil_command("MG _MT", axis=axis, expect_response=True)
        return resp

    def get_gearing_ratio(self, axis):
        """
        Return gearing ratio for given axis.

        """
        resp = float(self.galil_command(command=f'MG _GR{axis}', expect_response=True))
        return resp

    def get_gearing_lead(self, axis):
        """
        Return gearing lead for a given follower axis. A response of "0" means the axis is a leader.
        """
        resp = self.galil_command(command=f'GA{axis}=?', expect_response=True)

        return resp

    def set_motor_type(self, axis, motortype):
        """
        Set the motor type for each axis. The setting is typically 1, the servo motor (3-phased brushless)

        """
        resp = self.galil_command(command=f'MT{axis}={motortype};')
        return resp

    def get_off_on_error(self, axis):
        """
        Query the Off-On-Error (OE) state for an axis and return raw + human-readable.

        """
        resp = self.galil_command("MG _OE", axis=axis, expect_response=True)
        try:
            val = int(float(resp))
        except Exception:
            print(f"Unexpected response from MG _OE{axis}: {resp}")
            return None, "unknown"

        human_state = "enabled" if val == 1 else "disabled" if val == 0 else f"unknown ({val})"
        return val, human_state

    def set_off_on_error(self, axis, errtype):
        """
        Set the Off-On-Error (OE) function for the specified axis.

        """
        resp = self.galil_command(command=f'OE{axis}={errtype};')
        return resp

    def get_amp_gain(self, axis):
        """
        Query the amplifier gain (AG) value for an axis.

        """
        return self.galil_command("MG _AG", axis=axis, expect_response=True)

    def set_amp_gain(self, axis, val):
        """
        Set amplifier current/voltage gain for internal amplifier per axis.

        """
        resp = self.galil_command(command=f'AG{axis}={val};')
        return resp

    def get_torque_limit(self, axis):
        """
        Query the motor torque limit (TL) value for an axis.
        """
        return self.galil_command("MG _TL", axis=axis, expect_response=True)

    def set_torque_limit(self, axis, val):
        """
        Set motor torque limit per axis.

        """
        resp = self.galil_command(command=f'TL{axis}={val};')
        return resp

    def get_amp_currentloop_gain(self, axis):
        """
        Query the amplifier current loop gain (AU) value for an axis.

        """
        return self.galil_command("MG _AU", axis=axis, expect_response=True)

    def set_amp_currentloop_gain(self, axis, val):
        """
        Set amplifier current loop gain per axis.

        """
        resp = self.galil_command(command=f'AU{axis}={val};')
        return resp

    # initialization required if galil motor has sinusoidal amplifiers
    def enable_sin_commutation(self, axis):
        """
        For axes with a sinusoidal amplifier, the BA command is necessary to configure
        each axis for sinusoidal commutation

        """
        resp = self.galil_command(command=f'BA{axis};')
        return resp

    # initialization required if galil motor has sinusoidal amplifiers
    def set_magnetic_cycle(self, axis, val='3276.8'):
        """
        Defines the length of the motors magnetic cycle in encoder counts,
        required for correctly configuring sinusoidal commutation. Default is 3276.8

        """
        resp = self.galil_command(command=f'BM{axis}={val};')
        return resp

    # initialization required if galil motor has sinusoidal amplifiers
    def set_dwell_times(self, t_first, t_second):
        """
        Define the dwell times used for sinusoidal commutation initialization.

        This command sets the timing parameters for the Galil BZ command,
        specifying how long the motor will hold at the first and second
        magnetic positions during commutation (in milliseconds). These
        parameters must be defined before executing the BZ command.

        """
        t_first = int(round(float(t_first)))
        t_second = int(round(float(t_second)))
        cmd = f'BZ <{t_first}>{t_second}'
        resp = self.galil_command(command=cmd)
        return resp

    def initialize_axis(self, axis, val):
        """
        Initializes axes configured for sinusoidal commutation. BZ command
        will drive the motor to 2 different magnetic positions and then set
        the appropriate commutation angle. Cannot command with BZ unless BA
        and BM commands are sent first.

        """
        resp = self.galil_command(command=f'BZ{axis}={val};')
        return resp

    def define_position(self, axis, val=0):
        """
        Redefines current axis position to user specified value.
        Useful for homing procedure. Default is 0.

        """
        resp = self.galil_command(command=f'DP{axis}={val};')
        return resp

    def set_limitswitch_mode(self, axis, val):
        """
        Enable/disable limit switch detection on a given axis (LDx=val).

        val 0 = Disable no limit switches (all enabled)
        val 1 = Disable forward limit switch
        val 2 = Disable reverse limit switch
        val 3 = Disable both forward and reverse limit switches

        """
        resp = self.galil_command(command=f'LD{axis}={val};')
        return resp

    def get_limitswitch_mode(self, axis):
        """
        Returns the query from the disable limit switch command (LD) where:
        0 = Both limits enabled,
        1 = Forward disabled,
        2 = Reverse disabled,
        3 = Both forward and reverse disabled

        """
        resp = self.galil_command(command="MG _LD", axis=axis, expect_response=True)

        state = int(round(float(resp)))

        if state == 0:
            status = "Both limits enabled"
        if state == 1:
            status = "Forward disabled"
        if state == 2:
            status = "Reverse disabled"
        if state == 3:
            status = "Both forward and reverse disabled"

        return state, status

    def set_limitswitch_polarity(self, pol):
        """
        CN -1 means active low, CN +1 is active high. Default is  1
        (active high).

        """
        resp = self.galil_command(command=f'CN {pol};')
        return resp

    def get_limitswitch_polarity(self):
        """
        CN -1 means active low, CN +1 is active high. Our system is 1 (active high).

        """
        val = self.galil_command(command="MG _CN", expect_response=True)

        state = int(round(float(val)))

        if state == 1:
            status = "+1; active high"
        if state == -1:
            status = "-1, active low"

        return state, status

    def stop_motion(self, axis):
        """
        Stop motion.

        """
        cmd = f"ST {axis};"
        resp = self.galil_command(command=cmd)
        return resp

    def set_gearing(self, order):
        """
        Set gearing for defining follower and leader axes.
        The `order` string defines the sequence of operations.
        For example: `order = ',A,C'` means axis **B** follows
        **A**, and axis **D** follows **C** —the commas correspond
        to axes A, B, C, D in order.

        """
        resp = self.galil_command(command=f"GA {order};")
        return resp

    def set_gearing_ratio(self, order):
        """
        Set the gearing ratio for follower axes relative to their leader axes.
        A ratio of 1 means the follower moves at the same speed as its leader.
        For example, `GR 1,1` sets the gearing ratios for axes B and D

        """
        resp = self.galil_command(command=f"GR {order};")
        return resp

    def set_jog_speed(self, axis, speed):
        """
        Set jog speed for axis. Does not begin motion, just sets up speed
        for when ready to begin motion.

        """
        cmd = f"JG{axis}={speed};"
        resp = self.galil_command(command=cmd)
        return resp

    def set_speed(self, axis, speed):
        """
        Set speed for axis for use in commands like PR/PA (relative and absolute
        position). Does not begin motion, just defines speed for when
        ready to begin motion. Units of speed is in raw encoder units (counts/s).

        """
        cmd = f"SP{axis}={speed};"
        resp = self.galil_command(command=cmd)
        return resp

    def get_speed(self, axis):
        """
        Return speed for axis via MG _SP{axis}. Units of speed is in raw encoder units (counts/s).
        """
        resp = self.galil_command("MG _SP", axis=axis, expect_response=True)
        return resp

    def set_acceleration(self, axis, acc):
        """
        Set acceleration for axis. (encoder counts/sec^2) Value should be a power of 2. (e.g. 4096, 8192)
        """
        cmd = f"AC{axis}={acc};"
        resp = self.galil_command(command=cmd)
        return resp

    def get_acceleration(self, axis):
        """
        Return acceleration for given axis via "MG _AC{axis}". Units of acceleration are in raw encoder units (counts/s^2).
        """
        resp = self.galil_command("MG _AC", axis=axis, expect_response=True)
        return resp

    def set_deceleration(self, axis, decel):
        """
        Set deceleration for axis. Units are encoder counts/s^2
        """
        cmd = f"DC{axis}={decel};"
        resp = self.galil_command(command=cmd)
        return resp

    def get_deceleration(self, axis):
        """
        Return deceleration for given axis via "MG _DC{axis}". Units are in raw encoder units (counts/s^2).
        """
        resp = self.galil_command("MG _DC", axis=axis, expect_response=True)
        return resp

    def set_motor_state(self, axis, state):
        """
        Enable or disable a motor, then verify its state (0='on', 1='off').

        """
        state = state.lower().strip()
        if state not in ('enable', 'disable'):
            raise ValueError("state must be 'enable' or 'disable'.")

        # --- Execute command ---
        if state == 'enable':
            resp = self.galil_command("SH", axis=axis)
        elif state == 'disable':
            resp = self.galil_command("MO", axis=axis)
        return resp

    def get_motor_state(self, axis):
        """
        Query and interpret whether a motor is ON or OFF for a given axis.

        """
        resp = self.galil_command(command=f"MG _MO{axis}", expect_response=True)
        try:
            state = int(float(resp))
        except Exception:
            print(f"Unexpected response from MG _MO{axis}: {resp}")
            return None, "unknown"

        # --- Interpret the raw output ---
        human_state = "off" if state == 1 else "on" if state == 0 else f"unknown ({state})"
        return state, human_state



import argparse
import time

from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock

from socs.agents.hi6200.drivers import Hi6200Interface


class Hi6200Agent:
    """
    Agent to connect to the Hi6200 weight controller that measures the weight
    of the LN2 dewar on the SAT platform.

    Parameters:
        ip_address (string): IP address set on the Hi6200
        tcp_port (int): Modbus TCP port of the Hi6200.
                    Default set on the device is 502.
        scale (Hi6200Interface): A driver object that allows
                    for communication with the scale.
    """

    def __init__(self, agent, ip_address, tcp_port):
        self.agent = agent
        self.log = agent.log
        self.lock = TimeoutLock()

        self.ip_address = ip_address
        self.tcp_port = tcp_port
        self.scale = None

        self.monitor = False

        # Registers Scale Output
        agg_params = {
            'frame_length': 10 * 60,
        }
        self.agent.register_feed('scale_output',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('_')
    def init(self, session, params=None):
        """init()

        **Task** - Initialize connection to the Hi 6200 Weight Sensor.

        """
        with self.lock.acquire_timeout(0, job='init') as acquired:
            if not acquired:
                return False, "Could not acquire lock"

            self.scale = Hi6200Interface(self.ip_address, self.tcp_port)

            self.log.info("Connected to scale.")

        return True, 'Initialized Scale.'

    @ocs_agent.param('wait', type=float, default=1)
    def monitor_weight(self, session, params=None):
        """monitor_weight(wait=1)

        **Process** - Continuously monitor scale gross and net weights.

        Parameters:
            wait (float, optional): Time to wait between measurements
                [seconds].

        """
        self.monitor = True

        pm = Pacemaker(1, quantize=True)
        while self.monitor:

            pm.sleep()
            with self.lock.acquire_timeout(1, job='monitor_weight') as acquired:
                if not acquired:
                    self.log.warn("Could not start monitor_weight because "
                                  + f"{self.lock.job} is already running")
                    return False, "Could not acquire lock."

                data = {
                    'timestamp': time.time(),
                    'block_name': 'weight',
                    'data': {}
                }

                try:
                    # Grab the gross and net weights from the scale.
                    gross_weight = self.scale.read_scale_gross_weight()
                    net_weight = self.scale.read_scale_net_weight()

                    # The above functions return None when an Attribute error
                    # is thrown. If they did not return None and threw no
                    # errors, the data is good.
                    if (gross_weight is not None) and (net_weight is not None):
                        data['data']["Gross"] = gross_weight
                        data['data']["Net"] = net_weight
                        self.agent.publish_to_feed('scale_output', data)

                # Occurs when the scale disconnects.
                except AttributeError as e:
                    self.log.error("Connection with scale failed. Check that "
                                   + f"the scale is connected: {e}")
                    return False, "Monitoring weight failed"

                except ValueError as e:
                    self.log.error("Scale responded with an anomolous number, "
                                   + f"ignorning: {e}")

                except TypeError as e:
                    self.log.error("Scale responded with 'None' and broke the "
                                   + f"hex decoding, trying again: {e}")

        return True, "Finished monitoring weight"

    def stop_monitoring(self, session, params=None):
        self.monitor = False
        return True, "Stopping current monitor"


def make_parser(parser=None):
    """Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.

    """
    if parser is None:
        parser = argparse.ArgumentParser()

    # Add options specific to this agent.
    pgroup = parser.add_argument_group('Agent Options')
    pgroup.add_argument('--ip-address')
    pgroup.add_argument('--tcp-port')

    return parser


def main(args=None):

    parser = make_parser()
    args = site_config.parse_args(agent_class='Hi6200Agent',
                                  parser=parser,
                                  args=args)

    agent, runner = ocs_agent.init_site_agent(args)

    p = Hi6200Agent(agent, args.ip_address, int(args.tcp_port))

    agent.register_task('init', p.init)

    agent.register_process('monitor_weight', p.monitor_weight, p.stop_monitoring)

    runner.run(agent, auto_reconnect=True)


if __name__ == '__main__':
    main()



import argparse
import json
import os
import shutil
import time
from pathlib import Path

import requests
import txaio
import yaml
from ocs import ocs_agent, site_config
from ocs.ocs_twisted import Pacemaker, TimeoutLock
# Disable unverified HTTPS warnings (https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings)
from urllib3.exceptions import InsecureRequestWarning, ReadTimeoutError

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

# For logging
txaio.use_twisted()


class HTTPCameraAgent:
    """Grab screenshots from HTTP cameras.

    Parameters
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    config_file : str
        Config file path relative to OCS_CONFIG_DIR

    Attributes
    ----------
    agent : OCSAgent
        OCSAgent object which forms this Agent
    is_streaming : bool
        Tracks whether or not the agent is actively issuing requests to grab
        screenshots from cameras. Setting to false stops sending commands.
    log : txaio.tx.Logger
        txaio logger object, created by the OCSAgent
    """

    def __init__(self, agent, config_file, renew_token=2700):
        self.agent = agent
        self.is_streaming = False
        self.log = self.agent.log
        self.lock = TimeoutLock()

        file_path = os.path.join(os.environ['OCS_CONFIG_DIR'], config_file)
        with open(file_path, 'r') as f:
            self.config = yaml.safe_load(f)

        self.renew_token = renew_token

        agg_params = {
            'frame_length': 10 * 60  # [sec]
        }
        self.agent.register_feed('cameras',
                                 record=True,
                                 agg_params=agg_params,
                                 buffer_time=0)

    @ocs_agent.param('test_mode', default=False, type=bool)
    def acq(self, session, params=None):
        """acq(test_mode=False)

        **Process** - Grab screenshots from HTTP cameras.

        Parameters
        ----------
        test_mode : bool, optional
            Run the Process loop only once. Meant only for testing.
            Default is False.

        Notes
        -----
        The most recent data collected is stored in session.data in the
        structure::

            >>> response.session['data']
            # for each camera
            {'location1': {'location': 'location1',
                         'last_attempt': 1701983575.032506,
                         'connected': True,
                         'address': '10.10.10.41'},
             'location2': ...
            }
        """
        pm = Pacemaker(1 / 60, quantize=False)

        self.is_streaming = True
        while self.is_streaming:
            pm.sleep()
            # Use UTC
            timestamp = time.time()
            data = {}

            for camera in self.config['cameras']:
                data[camera['location']] = {'location': camera['location']}
                self.log.info(f"Grabbing screenshot from {camera['location']}")

                if camera['brand'] == 'reolink':
                    token = camera.get('token', None)
                    token_ts = camera.get('token_ts', 0)
                    # Token lease time is 1hr.
                    expired = (timestamp - token_ts) > self.renew_token
                    if token is None or expired:
                        login_url = f"https://{camera['address']}/api.cgi?cmd=Login"
                        login_payload = [{"cmd": "Login",
                                          "param": {"User":
                                                    {"Version": 0,
                                                     "userName": camera['user'],
                                                     "password": camera['password']}}}]
                        try:
                            resp = requests.post(login_url, data=json.dumps(login_payload), verify=False)
                        except (requests.exceptions.ConnectionError, requests.exceptions.RequestException, ReadTimeoutError) as e:
                            self.log.error(f'{e}')
                            self.log.info("Unable to get response from camera.")
                            data[camera['location']]['last_attempt'] = time.time()
                            data[camera['location']]['connected'] = False
                            continue
                        rdata = resp.json()
                        value = rdata[0].get('value', None)
                        if value is None:
                            self.log.info("Unable to get token. Max number of tokens used.")
                            data[camera['location']]['last_attempt'] = time.time()
                            data[camera['location']]['connected'] = False
                            continue
                        camera['token'] = value['Token']['name']
                        camera['token_ts'] = timestamp

                    payload = {'cmd': "Snap",
                               'channel': "0",
                               'rs': "flsYJfZgM6RTB_os",
                               'token': camera['token']}
                    url = f"https://{camera['address']}/cgi-bin/api.cgi"
                elif camera['brand'] == 'acti':
                    payload = {'USER': camera['user'],
                               'PWD': camera['password'],
                               'SNAPSHOT': camera.get('resolution', 'N640x480,100')}
                    url = f"http://{camera['address']}/cgi-bin/encoder"
                else:
                    self.log.info(f"{camera['brand']} is an unsupported camera brand. Skipping this config block.")
                    self.config['cameras'].remove(camera)
                    continue

                # Format directory and filename
                ctime = int(timestamp)
                ctime_dir = int(str(timestamp)[:5])
                Path(f"screenshots/{camera['location']}/{ctime_dir}").mkdir(parents=True, exist_ok=True)
                filename = f"screenshots/{camera['location']}/{ctime_dir}/{ctime}.jpg"
                latest_filename = f"screenshots/{camera['location']}/latest.jpg"

                # If no response from camera, update connection status and continue
                try:
                    if camera['brand'] == 'reolink':
                        response = requests.get(url, params=payload, stream=True, timeout=5, verify=False)
                    elif camera['brand'] == 'acti':
                        response = requests.get(url, params=payload, stream=True, timeout=5)
                except (requests.exceptions.ConnectionError, requests.exceptions.RequestException, ReadTimeoutError) as e:
                    self.log.error(f'{e}')
                    self.log.info("Unable to get response from camera.")
                    data[camera['location']]['last_attempt'] = time.time()
                    data[camera['location']]['connected'] = False
                    continue
                self.log.debug("Received screenshot from camera.")

                # Write screenshot to file and update latest file
                try:
                    with open(filename, 'wb') as out_file:
                        for chunk in response.iter_content(chunk_size=8192):
                            if chunk:
                                out_file.write(chunk)
                        out_file.flush()
                        os.fsync(out_file.fileno())
                    self.log.debug(f"Wrote {ctime}.jpg to /{camera['location']}/{ctime_dir}.")
                except (requests.exceptions.ConnectionError, requests.exceptions.RequestException, ReadTimeoutError) as e:
                    self.log.error(f'{e}')
                    self.log.info("Timeout occurred while writing to file.")
                    data[camera['location']]['last_attempt'] = time.time()
                    data[camera['location']]['connected'] = False
                    continue
                except Exception as e:
                    self.log.error(f'{e}')
                    self.log.info("Unexpected error occurred while writing to file.")
                    data[camera['location']]['last_attempt'] = time.time()
                    data[camera['location']]['connected'] = False
                    continue
                finally:
                    response.close()
                shutil.copy2(filename, latest_filename)
                self.log.debug(f"Updated latest.jpg in /{camera['location']}.")

                data[camera['location']]['last_attempt'] = time.time()
                data[camera['location']]['connected'] = True

            # Update session.data and publish to feed
            for camera in self.config['cameras']:
                data[camera['location']]['address'] = camera['address']
            session.data = data
            self.log.debug("{data}", data=session.data)

            message = {
                'block_name': 'cameras',
                'timestamp': timestamp,
                'data': {}
            }
            for camera in self.config['cameras']:
                message['data'][camera['location'] + "_connected"] = int(data[camera['location']]['connected'])
            session.app.publish_to_feed('cameras', message)
            self.log.debug("{msg}", msg=message)

            if params['test_mode']:
                break

        return True, "Finished Recording"

    def _stop_acq(self, session, params=None):
        """_stop_acq()
        **Task** - Stop task associated with acq process.
        """
        if self.is_streaming:
            session.set_status('stopping')
            self.is_streaming = False
            return True, "Stopping Recording"
        else:
            return False, "Acq is not currently running"


def add_agent_args(parser=None):
    """
    Build the argument parser for the Agent. Allows sphinx to automatically
    build documentation based on this function.
    """
    if parser is None:
        parser = argparse.ArgumentParser()

    pgroup = parser.add_argument_group("Agent Options")
    pgroup.add_argument("--config-file", type=str, help="Config file path relative to OCS_CONFIG_DIR")
    pgroup.add_argument("--mode", choices=['acq', 'test'])
    pgroup.add_argument("--renew-token", type=int, default=2700,
                        help="Renew API token after this amount of seconds. Used for Reolink cameras.")

    return parser


def main(args=None):
    # Start logging
    txaio.start_logging(level=os.environ.get("LOGLEVEL", "info"))

    parser = add_agent_args()
    args = site_config.parse_args(agent_class='HTTPCameraAgent',
                                  parser=parser,
                                  args=args)

    if args.mode == 'acq':
        init_params = True
    elif args.mode == 'test':
        init_params = False

    agent, runner = ocs_agent.init_site_agent(args)
    p = HTTPCameraAgent(agent,
                        config_file=args.config_file,
                        renew_token=args.renew_token)

    agent.register_process("acq",
                           p.acq,
                           p._stop_acq,
                           startup=init_params)

    runner.run(agent, auto_reconnect=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""blh motor driver controller"""

from datetime import datetime, timezone
from time import sleep

import numpy as np
import serial

from socs.agents.orientalmotor_blh.om_comm import (ALMCLR, INIT_CMD, MOVE_BKW,
                                                   MOVE_FWD, STATUS, STOP)

DEV_NAME = '/dev/ttyACM0'
UTC = timezone.utc


def calc_parity(data):
    """Parity calculation"""
    return np.bitwise_xor.reduce([i for i in data])


class BLH:
    """BLH motor driver controller"""

    def __init__(self, port=DEV_NAME):
        self._ser = serial.Serial(port, 9600, timeout=0.1)
        self._packet_index = 0
        self._start_dt = None
        self._connected = False

    def __w(self, data):
        self._ser.write(data)

    def __r(self, length):
        return self._ser.read(length)

    def connect(self):
        """Make connection to BLH"""
        self.__w(b'\x06')
        ret = self.__r(1)
        if ret[0] != 0x86:
            raise Exception('Not connected. Abort.')
        self._connected = True
        self._start_dt = datetime.now(tz=UTC)

        # Initialization codes
        for cmd in INIT_CMD:
            _ = self._wr(cmd)

    def _wr(self, data, readlen=40):
        """Write and read"""
        assert self._connected
        assert len(data) == 35
        tmpdt = datetime.now(tz=UTC)

        # Duration from the connection establishment
        ds_td = tmpdt - self._start_dt
        ds_int = int((ds_td.total_seconds() * 1e3) % 65536)
        ds_b = ds_int.to_bytes(2, 'little')

        # Packet index modulo 256
        pi_b = (self._packet_index % 256).to_bytes(1, 'little')

        # Packet creation
        pkt_pre = b'\xff' + data + ds_b + pi_b

        # The last byte is a parity
        parity = calc_parity(pkt_pre)
        pkt = pkt_pre + int(parity).to_bytes(1, 'little')

        self.__w(pkt)
        self._packet_index += 1

        return self.__r(readlen)

    def set_op_number(self, num):
        """Set operation number

        Parameter
        ---------
        num : int
            Operation number
        """
        assert 0 <= num < 16
        mark_0 = (((0b110 << 4) + num) << 5).to_bytes(2, 'little')
        mark_1 = (((num + 4) % 8) << 5) + (5 - int(num / 8))
        data = bytearray([0x03, 0x00, 0x01, 0x0a, 0x00, 0x86, 0x01]) + mark_0
        data += bytearray([0x04, 0x00, 0x00, mark_1]) + bytearray([0] * 22)

        return self._wr(data)

    def set_speed(self, speed):
        """Set speed

        Parameter
        ---------
        speed : int
            Speed in RPM
        """
        assert 50 <= speed <= 3000

        data = bytearray([0x03, 0x00, 0x01, 0x0c, 0x00, 0x81, 0x00, 0xc4, 0x01])
        data += speed.to_bytes(2, 'little')
        data += bytearray([0] * 3)

        # parity byte inside the packet.
        p_check = calc_parity(data[3:])
        data += bytearray([p_check] + [0] * 20)

        return self._wr(data)

    def set_accl_time(self, sec, accl=True):
        """Set acceleration time

        Parameters
        ----------
        sec : float
            Acceleration time in seconds

        accl : bool, default True
            True: Acceleration setting
            False: Decceleration setting
        """
        assert 0.5 <= sec <= 15.0

        data = bytearray([0x03, 0x00, 0x01, 0x0c, 0x00, 0x81, 0x00, 0xc5 if accl else 0xc6, 0x01])
        data += int(sec * 10).to_bytes(1, 'little')
        data += bytearray([0] * 4)

        # parity byte inside the packet
        p_check = calc_parity(data[3:])
        data += bytearray([p_check] + [0] * 20)
        return self._wr(data)

    def get_status(self):
        """Get status.

        Returns
        -------
        speed : int
            Speed in RPM
        """

        pkt_st = self._wr(STATUS, readlen=80)
        speed = int.from_bytes(pkt_st[7:11], 'little', signed=True)
        error = pkt_st[47]

        return speed, error

    def start(self, forward=True):
        """Start rotation.

        Parameter
        ---------
        forward : bool, default True
            Move forward if True

        Returns
        -------
        result : bool
            True if rotation command published correctly.
        """
        resp = self._wr(MOVE_FWD if forward else MOVE_BKW)
        result = resp[7] == 0

        return result

    def stop(self):
        """Stop rotation

        Returns
        -------
        result : bool
            True if stop command published correctly.
        """
        resp = self._wr(STOP)
        result = resp[7] == 0

        return result

    def clear_alarm(self):
        """Clear alarm"""
        self._wr(ALMCLR)


def main():
    """Test function"""
    blh = BLH()
    blh.connect()

    speed, error = blh.get_status()
    print(f'current speed: {speed}')
    print(f'current error: {error}')

    blh.set_speed(1500)
    blh.set_accl_time(5, False)
    blh.set_accl_time(5, True)

    print('MOVING FORWARD')
    blh.start(forward=True)
    for _ in range(200):
        print(f'{blh.get_status()}')
        sleep(0.1)

    print('STOPPED')
    blh.stop()
    for _ in range(10):
        print(f'{blh.get_status()}')
        sleep(0.1)

    print('MOVING BACKWARD')
    blh.start(forward=False)
    for _ in range(200):
        print(f'{blh.get_status()}')
        sleep(0.1)

    print('STOPPED')
    blh.stop()
    for _ in range(10):
        print(f'{blh.get_status()}')
        sleep(0.1)


if __name__ == '__main__':
    main()



import calendar
import datetime
import math
import pickle
import time
from dataclasses import dataclass, replace

import numpy as np

import socs.agents.acu.turnarounds as turnarounds

#: The number of seconds in a day.
DAY = 86400

#: Minimum number of points to group together on the start of a new
# leg, to not trigger programtrack error.
MIN_GROUP_NEW_LEG = 4

#: Registry for turn-around profile types.
TURNAROUNDS_ENUM = {
    'standard': 0,
    'standard_gen': 1,
    'three_leg': 2,
    'two_leg': 3,
}


def _progtrack_format_time(timestamp):
    fmt = '%j, %H:%M:%S'
    return (time.strftime(fmt, time.gmtime(timestamp))
            + '{:.6f}'.format(timestamp % 1.)[1:])


class FromFileScan:
    loop_time: float
    free_form: bool
    points: list
    preamble_count: int
    step_time: float
    az_range: tuple
    el_range: tuple


@dataclass
class TrackPoint:
    #: Timestamp of the point (unix timestamp)
    timestamp: float

    #: Azimuth (deg).
    az: float

    #: Elevation (deg).
    el: float

    #: Azimuth velocity (deg/s).
    az_vel: float

    #: Elevation velocity (deg/s).
    el_vel: float

    #: Az flag: 0 if stationary, 1 if non-final point of const-vel
    #: scan segment; 2 if final point of const-vel segment.
    az_flag: int = 0

    #: El flag: like az_flag but for el.
    el_flag: int = 0

    #: If 1, indicates that once this point is uploaded the next point
    #: in sequence also needs to be soon uploaded.  Used at start of a
    #: new const-vel scan segment.
    group_flag: int = 0


def track_point_time_shift(p, dt):
    return replace(p, timestamp=p.timestamp + dt)


def get_track_points_text(tpl, timestamp_offset=None, with_group_flag=False,
                          text_block=False):
    """Get a list of ProgramTrack lines for upload to ACU.

    Args:
      tpl (list): list of TrackPoint to convert.
      timestamp_offset (float): offset to add to all timestamps
        before rendering (defaults to 0).
      with_group_flag (bool): If True return each line as
        (group_flag, text).
      text_block (bool): If True, return all lines joined together
        into a single string.

    """
    if timestamp_offset is None:
        timestamp_offset = 0
    fmted_times = [_progtrack_format_time(p.timestamp + timestamp_offset)
                   for p in tpl]
    all_lines = [('{t}; {p.az:.6f}; {p.el:.6f}; {p.az_vel:.4f}; '
                  '{p.el_vel:.4f}; {p.az_flag}; {p.el_flag}\r\n')
                 .format(p=p, t=t)
                 for p, t in zip(tpl, fmted_times)]

    if text_block:
        return ''.join(all_lines)
    if with_group_flag:
        all_lines = [(p.group_flag, line) for p, line in zip(tpl, all_lines)]
    return all_lines


def from_file(filename, fmt=None):
    """Load a ProgramTrack trajectory from a file.  This function
    supports two formats. The modern format is a pickle file. The
    older numpy format is also supported.

    Parameters:
      filename (str): Full path to the file.
      fmt (str): Optional, one of "pickle" or "numpy". If this is
        unspecified, the code will assume pickle format unless the
        filename ends in "npy".

    Returns:
      FromFileScan: Object containing the loaded points and
        supporting config for ProgramTrack mode.

    Notes:

      For the pickle-based file format, the file must encode a
      single dict. Here is a minimal example::

        {
         'timestamp': [1747233498, 1747233499, ..., 1747233523],
         'az': [180, 181, ..., 160],
         'el': [60, 60, ..., 60],
        }

      Those 3 required entries are "vectors", with the same
      length. The vectors may be any iterable type but please use
      lists or ndarrays. Optionally, the user may specify these other
      vectors:

        - 'az_vel': the az velocity at each point.
        - 'el_vel': the el velocity at each point.
        - 'az_flag': the az flag.
        - 'el_flag': the el flag.
        - 'group_flag': the point grouping flag.

      If not provided, the "vel" vectors will be computed from the
      gradient of the position vectors.  The "flag" vectors will
      default to all 0.  See :class:`TrackPoint
      <socs.agents.acu.drivers.TrackPoint>` for purpose of the various
      flags.

      The following settings (stated with their default values) may
      also be included in the dict:

        - 'free_form' (bool, False): Specifies whether the track
          should be run with the turn-around profiler disabled and
          spline (rather than linear) interpolation enabled. When
          false, the ACU will do linear interpolation *and* automatic
          profiling of turn-arounds, between constant velocity
          segments. Setting free_form=False is appropriate for
          constant elevation, constant az speed scans.  Set
          free_form=True for more complex scans.
        - 'loopable' (bool, False): Specifies whether the track should
          be repeated, forever.  When this is the case, the final
          point of the track (i.e. the last point in all the vectors)
          is ignored *except for the timestamp*, which is used to set
          the time at which the next loop iteration is started.
        - 'preamble_count' (int, 0): If loopable, this specifies the
          number of points in the track (i.e. in each vector) that are
          not included in the loopable portion.  This permits the
          track to have a ramp-up, or partial initial scan segment,
          prior to entering the repetable template.

      The older numpy-based format does not support the additional
      settings.  The numpy file must contain an iterable with 5 or 7
      entries, where all entries are 1-d arrays of the same length.
      The first 5 arrays will correspond to 'timestamp', 'az', 'el',
      'az_vel', 'el_vel'.  The 2 optional arrays are 'az_flag' and
      'el_flag'.

    """
    if fmt is None:
        fmt = 'pickle'
        if filename.endswith('npy'):
            fmt = 'numpy'

    if fmt == 'numpy':
        info = np.load(filename)
        if len(info) not in [5, 7]:
            raise ValueError(f'Unexpected field count ({len(info)}) in {filename}')
        times, az, el, vaz, vel = info[:5]
        if len(info) == 5:
            az_flags = np.zeros(len(times), int)
            el_flags = az_flags
        elif len(info) == 7:
            az_flags = info[5].astype('int')
            el_flags = info[6].astype('int')

        output = FromFileScan()
        output.loop_time = 0.
        output.free_form = False
        output.step_time = np.diff(times).min()
        output.points = [TrackPoint(*a) for a in zip(
            times, az, el, vaz, vel, az_flags, el_flags)]
        output.az_range = (az.min(), az.max())
        output.el_range = (el.min(), el.max())

    elif fmt == 'pickle':
        data = pickle.load(open(filename, 'rb'))
        output = FromFileScan()
        output.loop_time = 0.
        output.free_form = data.get('free_form', False)

        keys = ['timestamp', 'az', 'el', 'az_vel', 'el_vel',
                'az_flag', 'el_flag', 'group_flag']
        vects = {k: data.get(k) for k in keys}
        n = len(vects['az'])

        dtv = np.gradient(vects['timestamp'])
        if vects['az_vel'] is None:
            vects['az_vel'] = np.gradient(vects['az']) / dtv
        if vects['el_vel'] is None:
            vects['el_vel'] = np.gradient(vects['el']) / dtv
        if vects['az_flag'] is None:
            vects['az_flag'] = np.zeros(n, int)
        if vects['el_flag'] is None:
            vects['el_flag'] = np.zeros(n, int)
        if vects['group_flag'] is None:
            vects['group_flag'] = np.zeros(n, int)

        output.preamble_count = data.get('preamble_count', 0)
        if data.get('loopable'):
            # Measure repeat time.
            output.loop_time = (vects['timestamp'][-1]
                                - vects['timestamp'][output.preamble_count])
            # ... and drop last point.
            for k in keys:
                vects[k] = vects[k][:-1]

        columns = [vects[k] for k in keys]
        output.points = [TrackPoint(*row) for row in zip(*columns)]

        output.step_time = np.diff(vects['timestamp']).min()
        output.az_range = (vects['az'].min(), vects['az'].max())
        output.el_range = (vects['el'].min(), vects['el'].max())

    else:
        raise ValueError(f"Invalid fmt={fmt}")

    return output


def timecode(acutime, now=None):
    """Takes the time code produced by the ACU status stream and returns
    a unix timestamp.

    Parameters:
        acutime (float): The time recorded by the ACU status stream,
            corresponding to the fractional day of the year.
        now (float): The time, as unix timestamp, to assume it is now.
            This is for testing, it defaults to time.time().

    """
    sec_of_day = (acutime - 1) * DAY
    if now is None:
        now = time.time()  # testing

    # This guard protects us at end of year, when time.time() and
    # acutime might correspond to different years.
    if acutime > 180:
        context = datetime.datetime.utcfromtimestamp(now - 30 * DAY)
    else:
        context = datetime.datetime.utcfromtimestamp(now + 30 * DAY)

    year = context.year
    gyear = calendar.timegm(time.strptime(str(year), '%Y'))
    comptime = gyear + sec_of_day
    return comptime


def _get_target_az(current_az, current_t, increasing, az_endpoint1, az_endpoint2, az_speed, az_drift):
    # Return the next endpoint azimuth, based on current (az, t)
    # and whether to move in +ve or -ve az direction.
    #
    # Includes the effects of az_drift, to keep the scan endpoints
    # (at least at the end of a scan) on the drifted trajectories.
    if increasing:
        target = max(az_endpoint1, az_endpoint2)
    else:
        target = min(az_endpoint1, az_endpoint2)
    if az_drift is not None:
        v = az_speed if increasing else -az_speed
        target = target + az_drift / (v - az_drift) * (
            (target - current_az + v * current_t))
    return target


def generate_constant_velocity_scan(az_endpoint1, az_endpoint2, az_speed,
                                    acc, el_endpoint1, el_endpoint2,
                                    el_speed=0,
                                    num_batches=None,
                                    num_scans=None,
                                    start_time=None,
                                    wait_to_start=10.,
                                    step_time=1.,
                                    batch_size=500,
                                    az_start='mid_inc',
                                    az_first_pos=None,
                                    az_drift=None,
                                    turnaround_method='standard'):
    """Python generator to produce times, azimuth and elevation positions,
    azimuth and elevation velocities, azimuth and elevation flags for
    arbitrarily long constant-velocity azimuth scans.

    Parameters:
        az_endpoint1 (float): azimuth endpoint for the scan start
        az_endpoint2 (float): second azimuth endpoint of the scan
        az_speed (float): speed of the constant-velocity azimuth motion
        acc (float): turnaround acceleration for the azimuth motion at the
            endpoints
        el_endpoint1 (float): elevation endpoint for the scan start
        el_endpoint2 (float): second elevation endpoint of the scan. For
            constant az scans, this must be equal to el_endpoint1.
        el_speed (float): speed of the elevation motion. For constant az
            scans, set to 0.0
        num_batches (int or None): sets the number of batches for the
            generator to create. Default value is None (interpreted as infinite
            batches).
        num_scans (int or None): if not None, limits the points
          returned to the specified number of constant velocity legs.
        start_time (float or None): a ctime at which to start the scan.
            Default is None, which is interpreted as starting now +
            wait_to_start.
        wait_to_start (float): number of seconds to wait between
            start_time and when the scan actually starts. Default is 10 seconds.
        step_time (float): time between points on the constant-velocity
            parts of the motion. Default value is 1.0 seconds. Minimum value is
            0.05 seconds.
        batch_size (int): number of values to produce in each iteration.
            Default is 500. Batch size is reset to the length of one leg of the
            motion if num_batches is not None.
        az_start (str): part of the scan to start at.  To start at one
            of the extremes, use 'az_endpoint1', 'az_endpoint2', or
            'end' (same as 'az_endpoint1').  To start in the midpoint
            of the scan use 'mid_inc' (for first half-leg to have
            positive az velocity), 'mid_dec' (negative az velocity),
            or 'mid' (velocity oriented towards endpoint2).
        az_first_pos (float): If not None, the first az scan will
            start at this position (but otherwise proceed in the same
            starting direction).
        az_drift (float): The rate (deg / s) at which to shift the
            scan endpoints in time.  This can be used to better track
            celestial sources in targeted scans.
        turnaround_method (str): The method used for generating turnaround.
            (Default) 'standard' generates the baseline minimal jerk trajectory.
            'standard_gen' generates a list of track_points throughout the turnaround
            to mimic the standard acu turnaround.
            'three_leg' generates a three-leg turnaround which attempts to
            minimize the acceleration at the midpoint of the turnaround.
            'two_leg' generates a "three_leg" turnaround with second_leg_time = 0.

    Yields:
        points (list): a list of TrackPoint objects.  Raises
          StopIteration once exit condition, if defined, is met.

    """
    if az_endpoint1 == az_endpoint2:
        raise ValueError('Generator requires two different az endpoints!')

    # Force the el_speed to 0.  It matters because an el_speed in
    # ProgramTrack data that exceeds the ACU limits will cause the
    # point to be rejected, even if there's no motion in el planned
    # (which, at the time of this writing, there is not).
    el_speed = 0.

    # Note that starting scan direction gets modified, below,
    # depending on az_start.
    increasing = az_endpoint2 > az_endpoint1

    if az_start in ['az_endpoint1', 'az_endpoint2', 'end']:
        if az_start in ['az_endpoint1', 'end']:
            az = az_endpoint1
        else:
            az = az_endpoint2
            increasing = not increasing
    elif az_start in ['mid_inc', 'mid_dec', 'mid']:
        az = (az_endpoint1 + az_endpoint2) / 2
        if az_start == 'mid':
            pass
        elif az_start == 'mid_inc':
            increasing = True
        else:
            increasing = False
    else:
        raise ValueError(f'az_start value "{az_start}" not supported. Choose from '
                         'az_endpoint1, az_endpoint2, mid_inc, mid_dec')
    az_vel = az_speed if increasing else -az_speed

    # Bias the starting point for the first leg?
    if az_first_pos is not None:
        az = az_first_pos

    if start_time is None:
        t0 = time.time() + wait_to_start
    else:
        t0 = start_time
    t = 0
    turntime = 2.0 * az_speed / acc
    el = el_endpoint1
    if step_time < 0.05:
        raise ValueError('Time step size too small, must be at least '
                         '0.05 seconds')
    daz = step_time * az_speed
    el_vel = el_speed
    az_flag = 0
    el_flag = 0
    if num_batches is None:
        stop_iter = float('inf')
    else:
        stop_iter = num_batches
        batch_size = int(np.ceil(abs(az_endpoint2 - az_endpoint1) / daz))

    def dec_num_scans():
        nonlocal num_scans
        if num_scans is not None:
            num_scans -= 1

    def check_num_scans():
        return num_scans is None or num_scans > 0

    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed, az_drift)
    point_group_batch = 0

    i = 0
    point_queue = []
    while i < stop_iter and check_num_scans():
        i += 1
        point_block = []
        for j in range(batch_size):
            if len(point_queue):  # Pull from points in the queue first
                point_block.append(point_queue.pop(0))
                continue

            point_block.append(TrackPoint(
                timestamp=t + t0,
                az=az, el=el, az_vel=az_vel, el_vel=el_vel,
                az_flag=az_flag, el_flag=el_flag,
                group_flag=int(point_group_batch > 0)))

            if point_group_batch > 0:
                point_group_batch -= 1

            if increasing:
                if az <= (target_az - 2 * daz):
                    t += step_time
                    az += daz
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                elif az == target_az:
                    # Turn around.
                    turnaround_track = turnarounds.gen_turnaround(turnaround_method=turnaround_method,
                                                                  t0=t + t0, az0=az, el0=el, v0=az_vel,
                                                                  turntime=turntime,
                                                                  az_flag=az_flag, el_flag=el_flag,
                                                                  point_group_batch=point_group_batch)
                    point_queue.extend(turnaround_track)

                    t += turntime
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = False
                    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed, az_drift)
                    dec_num_scans()
                    point_group_batch = MIN_GROUP_NEW_LEG - 1
                else:
                    time_remaining = (target_az - az) / az_speed
                    az = target_az
                    t += time_remaining
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0
            else:
                if az >= (target_az + 2 * daz):
                    t += step_time
                    az -= daz
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                elif az == target_az:
                    # Turn around.
                    turnaround_track = turnarounds.gen_turnaround(turnaround_method=turnaround_method,
                                                                  t0=t + t0, az0=az, el0=el, v0=az_vel,
                                                                  turntime=turntime,
                                                                  az_flag=az_flag, el_flag=el_flag,
                                                                  point_group_batch=point_group_batch)
                    point_queue.extend(turnaround_track)

                    t += turntime
                    az_vel = az_speed
                    el_vel = el_speed
                    az_flag = 1
                    el_flag = 0
                    increasing = True
                    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed, az_drift)
                    dec_num_scans()
                    point_group_batch = MIN_GROUP_NEW_LEG - 1
                else:
                    time_remaining = (az - target_az) / az_speed
                    az = target_az
                    t += time_remaining
                    az_vel = -1 * az_speed
                    el_vel = el_speed
                    az_flag = 2
                    el_flag = 0

            if not check_num_scans():
                # Kill the velocity on the last point and exit -- this
                # was recommended at LAT FAT for smoothly stopping the
                # motion at end of program.
                point_block[-1].az_vel = 0
                point_block[-1].el_vel = 0
                break

        yield point_block


def generate_type3_scan(az_endpoint1, az_endpoint2, az_speed,
                        acc, el_endpoint1, el_endpoint2,
                        el_freq=.15,
                        az_vel_ref=None,
                        num_batches=None,
                        num_scans=None,
                        start_time=None,
                        wait_to_start=10.,
                        step_time=1.,
                        batch_size=500,
                        az_start='mid_inc',
                        az_first_pos=None,
                        az_drift=None,
                        turnaround_method='two_leg'):
    """Python generator to produce times, azimuth and elevation positions,
    azimuth and elevation velocities, azimuth and elevation flags for
    arbitrarily long type 3 scan.

    Parameters:
        az_endpoint1 (float): azimuth endpoint for the scan start
        az_endpoint2 (float): second azimuth endpoint of the scan
        az_speed (float): speed of the constant-velocity azimuth motion
        acc (float): turnaround acceleration for the azimuth motion at the
            endpoints
        el_endpoint1 (float): elevation endpoint for the scan start
        el_endpoint2 (float): second elevation endpoint of the scan. For
            constant az scans, this must be equal to el_endpoint1.
        el_freq(float): frequency of the elevation nods in Hz.
        az_vel_ref(float or None): azimuth to center the velocity profile at.
                                   If None then the average of the endpoints is used.
        num_batches (int or None): sets the number of batches for the
            generator to create. Default value is None (interpreted as infinite
            batches).
        num_scans (int or None): if not None, limits the points
          returned to the specified number of constant velocity legs.
        start_time (float or None): a ctime at which to start the scan.
            Default is None, which is interpreted as starting now +
            wait_to_start.
        wait_to_start (float): number of seconds to wait between
            start_time and when the scan actually starts. Default is 10 seconds.
        step_time (float): time between points on the constant-velocity
            parts of the motion. Default value is 1.0 seconds. Minimum value is
            0.05 seconds.
        batch_size (int): number of values to produce in each iteration.
            Default is 500. Batch size is reset to the length of one leg of the
            motion if num_batches is not None.
        az_start (str): part of the scan to start at.  To start at one
            of the extremes, use 'az_endpoint1', 'az_endpoint2', or
            'end' (same as 'az_endpoint1').  To start in the midpoint
            of the scan use 'mid_inc' (for first half-leg to have
            positive az velocity), 'mid_dec' (negative az velocity),
            or 'mid' (velocity oriented towards endpoint2).
        az_first_pos (float): If not None, the first az scan will
            start at this position (but otherwise proceed in the same
            starting direction).
        az_drift (float): The rate (deg / s) at which to shift the
            scan endpoints in time.  This can be used to better track
            celestial sources in targeted scans.
        turnaround_method (str): The method used for generating turnaround.
            'standard' is unusable with type3 scans!
            'standard_gen' generates a list of track_points throughout the turnaround
            to mimic the standard acu turnaround and is usable for type3 scans.
            'three_leg' generates a three-leg turnaround which attempts to
            minimize the acceleration at the midpoint of the turnaround.
            (Default) 'two_leg' generates a "three_leg" turnaround with second_leg_time = 0.

    Yields:
        points (list): a list of TrackPoint objects.  Raises
          StopIteration once exit condition, if defined, is met.

    """
    def get_scan_time(az0, az1, az_speed, az_cent):
        upper = -1 * np.cos(np.deg2rad(az1 - az_cent))
        lower = -1 * np.cos(np.deg2rad(az0 - az_cent))

        return abs(upper - lower) / np.deg2rad(az_speed)

    if az_endpoint1 == az_endpoint2:
        raise ValueError('Generator requires two different az endpoints!')

    if az_drift is not None:
        raise ValueError("Az drift not supported for type 2 or 3 scans!")

    # Get center of az range
    if az_vel_ref is None:
        az_vel_ref = (az_endpoint1 + az_endpoint2) / 2.
    az_cent = az_vel_ref - 90

    if any([abs(_az - az_vel_ref) > 70.
            for _az in [az_endpoint1, az_endpoint2]]):
        raise ValueError("Az limits for type 2 and 3 scans must not be more than 70 "
                         "degrees away from az_vel_ref.")

    # Get el throw
    el_throw = abs(el_endpoint2 - el_endpoint1) / 2
    el_cent = (el_endpoint1 + el_endpoint2) / 2.

    # Note that starting scan direction gets modified, below,
    # depending on az_start.
    increasing = az_endpoint2 > az_endpoint1

    if az_start in ['az_endpoint1', 'az_endpoint2', 'end']:
        if az_start in ['az_endpoint1', 'end']:
            az = az_endpoint1
        else:
            az = az_endpoint2
            increasing = not increasing
    elif az_start in ['mid_inc', 'mid_dec', 'mid']:
        az = (az_endpoint1 + az_endpoint2) / 2
        if az_start == 'mid':
            pass
        elif az_start == 'mid_inc':
            increasing = True
        else:
            increasing = False
    else:
        raise ValueError(f'az_start value "{az_start}" not supported. Choose from '
                         'az_endpoint1, az_endpoint2, mid_inc, mid_dec')
    az_vel = az_speed if increasing else -az_speed

    # Bias the starting point for the first leg?
    if az_first_pos is not None:
        az = az_first_pos

    if start_time is None:
        t0 = time.time() + wait_to_start
    else:
        t0 = start_time

    vel_0 = az_speed / np.sin(np.deg2rad(az_endpoint1 - az_cent))
    vel_1 = az_speed / np.sin(np.deg2rad(az_endpoint2 - az_cent))
    min_tt = {1: (0.85 * abs(vel_0) / 9 * 11.616)**.5, -1: (0.85 * abs(vel_1) / 9 * 11.616)**.5}
    tt = {1: max(2 * vel_0 / acc, min_tt[1]), -1: max(2 * vel_1 / acc, min_tt[-1])}
    t = 0
    el = el_endpoint1
    if step_time < 0.05:
        raise ValueError('Time step size too small, must be at least '
                         '0.05 seconds')
    az_flag = 0
    el_flag = 0
    if num_batches is None:
        stop_iter = float('inf')
    else:
        stop_iter = num_batches
        batch_size = int(np.ceil(get_scan_time(az_endpoint1, az_endpoint2, az_speed, az_cent) / step_time))

    def dec_num_scans():
        nonlocal num_scans
        if num_scans is not None:
            num_scans -= 1

    def check_num_scans():
        return num_scans is None or num_scans > 0

    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed / np.sin(np.deg2rad(az - az_cent)), az_drift)
    point_group_batch = 0

    def get_el(_t):
        return (el_cent - el_throw * np.cos(_t * el_freq * 2 * np.pi),
                el_throw * el_freq * 2 * np.pi * np.sin(_t * el_freq * 2 * np.pi))

    i = 0
    point_queue = []
    while i < stop_iter and check_num_scans():
        i += 1
        point_block = []
        for j in range(batch_size):
            if len(point_queue):  # Pull from points in the queue first
                point_block.append(point_queue.pop(0))
                continue

            point_block.append(TrackPoint(
                timestamp=t + t0,
                az=az, el=el_cent, az_vel=az_vel / np.sin(np.deg2rad(az - az_cent)), el_vel=0,
                az_flag=az_flag, el_flag=el_flag,
                group_flag=int(point_group_batch > 0)))

            if point_group_batch > 0:
                point_group_batch -= 1

            if increasing:
                if get_scan_time(az, target_az, az_speed, az_cent) > 2 * step_time:
                    t += step_time
                    az += step_time * az_speed / np.sin(np.deg2rad(az - az_cent))
                    az_vel = az_speed
                    az_flag = 1  # 1
                    el_flag = 0
                elif az == target_az:
                    point_group_batch = MIN_GROUP_NEW_LEG - 1
                    # Turn around.
                    _v = az_vel / np.sin(np.deg2rad(az - az_cent))
                    turnaround_track = turnarounds.gen_turnaround(turnaround_method=turnaround_method,
                                                                  t0=t + t0, az0=az, el0=el, v0=_v,
                                                                  turntime=tt[1],
                                                                  az_flag=az_flag, el_flag=el_flag,
                                                                  step_time=step_time,
                                                                  point_group_batch=point_group_batch)
                    point_queue.extend(turnaround_track)

                    # Turn around.
                    t += tt[1]
                    az_vel = -1 * az_speed
                    az_flag = 1  # 1
                    el_flag = 0
                    increasing = False
                    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed / np.sin(np.deg2rad(az - az_cent)), az_drift)
                    dec_num_scans()
                else:
                    time_remaining = get_scan_time(az, target_az, az_speed, az_cent)
                    az = target_az
                    t += time_remaining
                    az_vel = az_speed
                    az_flag = 1  # 2
                    el_flag = 0
            else:
                if get_scan_time(az, target_az, az_speed, az_cent) > 2 * step_time:
                    t += step_time
                    az -= step_time * az_speed / np.sin(np.deg2rad(az - az_cent))
                    az_vel = -1 * az_speed
                    az_flag = 1  # 1
                    el_flag = 0
                elif az == target_az:
                    point_group_batch = MIN_GROUP_NEW_LEG - 1
                    # Turn around.
                    _v = az_vel / np.sin(np.deg2rad(az - az_cent))
                    turnaround_track = turnarounds.gen_turnaround(turnaround_method=turnaround_method,
                                                                  t0=t + t0, az0=az, el0=el, v0=_v,
                                                                  turntime=tt[1],
                                                                  az_flag=az_flag, el_flag=el_flag,
                                                                  step_time=step_time,
                                                                  point_group_batch=point_group_batch)
                    point_queue.extend(turnaround_track)

                    # Turn around.
                    t += tt[-1]
                    az_vel = az_speed
                    az_flag = 1  # 1
                    el_flag = 0
                    increasing = True
                    target_az = _get_target_az(az, t, increasing, az_endpoint1, az_endpoint2, az_speed / np.sin(np.deg2rad(az - az_cent)), az_drift)
                    dec_num_scans()
                else:
                    time_remaining = get_scan_time(az, target_az, az_speed, az_cent)
                    az = target_az
                    t += time_remaining
                    az_vel = -1 * az_speed
                    az_flag = 1  # 2
                    el_flag = 0

            if not check_num_scans():
                # Kill the velocity on the last point and exit -- this
                # was recommended at LAT FAT for smoothly stopping the
                # motion at end of program.
                point_block[-1].az_vel = 0
                point_block[-1].el_vel = 1000
                break

        for p in point_block:
            if p.el_vel == 1000:
                p.el_vel = 0.
            elif el_throw != 0:
                p.el, p.el_vel = get_el(p.timestamp - t0)
                p.el_flag = 1

        yield point_block


def generate_type2_scan(az_endpoint1, az_endpoint2, az_speed,
                        acc, el_endpoint1,
                        az_vel_ref=None,
                        num_batches=None,
                        num_scans=None,
                        start_time=None,
                        wait_to_start=10.,
                        step_time=1.,
                        batch_size=500,
                        az_start='mid_inc',
                        az_first_pos=None,
                        az_drift=None,
                        turnaround_method='two_leg'):
    """Python generator to produce times, azimuth and elevation positions,
    azimuth and elevation velocities, azimuth and elevation flags for
    arbitrarily long type 2 scan.

    Parameters:
        az_endpoint1 (float): azimuth endpoint for the scan start
        az_endpoint2 (float): second azimuth endpoint of the scan
        az_speed (float): speed of the constant-velocity azimuth motion
        acc (float): turnaround acceleration for the azimuth motion at the
            endpoints
        el_endpoint1 (float): elevation endpoint for the scan start
        az_vel_ref(float or None): azimuth to center the velocity profile at.
                                   If None then the average of the endpoints is used.
        num_batches (int or None): sets the number of batches for the
            generator to create. Default value is None (interpreted as infinite
            batches).
        num_scans (int or None): if not None, limits the points
          returned to the specified number of constant velocity legs.
        start_time (float or None): a ctime at which to start the scan.
            Default is None, which is interpreted as starting now +
            wait_to_start.
        wait_to_start (float): number of seconds to wait between
            start_time and when the scan actually starts. Default is 10 seconds.
        step_time (float): time between points on the constant-velocity
            parts of the motion. Default value is 1.0 seconds. Minimum value is
            0.05 seconds.
        batch_size (int): number of values to produce in each iteration.
            Default is 500. Batch size is reset to the length of one leg of the
            motion if num_batches is not None.
        az_start (str): part of the scan to start at.  To start at one
            of the extremes, use 'az_endpoint1', 'az_endpoint2', or
            'end' (same as 'az_endpoint1').  To start in the midpoint
            of the scan use 'mid_inc' (for first half-leg to have
            positive az velocity), 'mid_dec' (negative az velocity),
            or 'mid' (velocity oriented towards endpoint2).
        az_first_pos (float): If not None, the first az scan will
            start at this position (but otherwise proceed in the same
            starting direction).
        az_drift (float): The rate (deg / s) at which to shift the
            scan endpoints in time.  This can be used to better track
            celestial sources in targeted scans.
        turnaround_method (str): The method used for generating turnaround.
            'standard' is unusable with type2 scans!
            'standard_gen' generates a list of track_points throughout the turnaround
            to mimic the standard acu turnaround and is usable for type2 scans.
            'three_leg' generates a three-leg turnaround which attempts to
            minimize the acceleration at the midpoint of the turnaround.
            'two_leg' (Default) generates a "three_leg" turnaround with second_leg_time = 0.

    Yields:
        points (list): a list of TrackPoint objects.  Raises
          StopIteration once exit condition, if defined, is met.

    """
    return generate_type3_scan(az_endpoint1, az_endpoint2, az_speed,
                               acc, el_endpoint1, el_endpoint1,
                               el_freq=0,
                               az_vel_ref=az_vel_ref,
                               num_batches=num_batches,
                               num_scans=num_scans,
                               start_time=start_time,
                               wait_to_start=wait_to_start,
                               step_time=step_time,
                               batch_size=batch_size,
                               az_start=az_start,
                               az_first_pos=az_first_pos,
                               az_drift=az_drift,
                               turnaround_method=turnaround_method)


def plan_scan(az_end1, az_end2, el, v_az=1, a_az=1, az_start=None,
              scan_type=1):
    """Determine some important parameters for running a ProgramTrack
    scan with the desired end points, velocity, and mean turn-around
    acceleration.

    These get complicated in the limit of high velocity and narrow
    scan.

    Returns:

      A dict with outputs of the calculations.  The following items
      must be considered when generating and posting the track points:

      - 'step_time': The recommended track point separation, in
        seconds.
      - 'wait_to_start': The minimum time (s) between initiating
        ProgramTrack mode and the first uploaded point's timestamp.
      - 'init_az': The az (deg) at which to position the telescope
        before beginning the scan.  This takes into account any "ramp
        up" that needs to occur and the fact that such ramp up needs
        to be finished before the ACU starts profiling the first
        turn-around.

      The following dict items provide additional detail /
      intermediate results:

      - 'scan_start_buffer': Minimum amount (deg of az) by which to
        shift the start of the first scan leg in order to satisfy the
        requirements for az_prep and az_rampup.  This ultimately is
        what can make init_az different from the natural first leg
        starting point.  This parameter is always non-negative.
      - 'turnprep_buffer': Minimum azimuth travel required for
        ProgramTrack to prepare a turn-around.
      - 'rampup_buffer': Minimum azimuth travel required for
        ProgramTrack to ramp up to the first leg velocity.  Degrees,
        positive.
      - 'rampup_time': Number of seconds before the first track point
        where the platform could start moving (as part of smooth
        acceleration into the initial velocity).

    """
    # Convert Agent-friendly arguments to az/throw/init
    az = (az_end1 + az_end2) / 2
    throw = (az_end2 - az_end1) / 2

    if az_start in [None, 'mid']:
        init = 'mid'
    elif az_start == 'mid_inc':
        init = 'mid'
        throw = abs(throw)
    elif az_start == 'mid_dec':
        init = 'mid'
        throw = -abs(throw)
    elif az_start in ['az_endpoint1', 'end']:
        init = 'end'
    elif az_start in ['az_endpoint2']:
        init = 'end'
        throw = -throw
    else:
        raise ValueError(f'Unexpected az_start={az_start}')

    # Info to pass back.
    plan = {}

    # Point time separation: at least 5 points per leg, preferably 10.
    if scan_type in [2, 3]:
        dt = .05
    else:
        dt = 2 * abs(throw / v_az) / 10
        dt = min(max(dt, 0.1), 1.0)
    assert (2 * abs(throw / v_az) / dt >= 5)
    plan['step_time'] = dt

    # Turn around prep distance (deg)? 5 point periods, times the vel.
    turnprep_buffer = 5 * dt * v_az

    # Ramp-up distance needed
    a0 = 1.  # Peak accel of ramp-up...
    rampup_buffer = v_az**2 / a0
    plan['turnprep_buffer'] = turnprep_buffer
    plan['rampup_buffer'] = rampup_buffer

    # Any az ramp-up prep required?
    if init == 'mid':
        scan_start_buffer = max(turnprep_buffer + rampup_buffer - abs(throw), 0)
    elif init == 'end':
        scan_start_buffer = max(turnprep_buffer + rampup_buffer - 2 * abs(throw), 0)
    plan['scan_start_buffer'] = scan_start_buffer

    # Set wait time (this comes out a little lower than its supposed to...)
    # plan['wait_time'] = v_az / a0 * 2
    plan['rampup_time'] = v_az / a0
    plan['wait_to_start'] = max(5, plan['rampup_time'] * 1.2)

    # Fill out some other useful info...
    plan['init_az'] = az - math.copysign(scan_start_buffer, throw)
    if init == 'end':
        plan['init_az'] -= throw

    return plan



#!/usr/bin/env python3
"""dS378 ethernet relay"""
import sys
from enum import IntEnum

from socs.tcp import TCPInterface

# Command byte, length of expected response
GET_STATUS = 0x30, 8
SET_RELAY = 0x31, 1
SET_OUTPUT = 0x32, 1
GET_RELAYS = 0x33, 5
GET_INPUTS = 0x34, 2
GET_ANALOG = 0x35, 14
GET_COUNTERS = 0x36, 8

# Number of trial for send / receive
N_TRIAL = 2
LEN_BUFF = 1024


class RelayStatus(IntEnum):
    """Relay status"""
    on = 1
    off = 0


class DS378(TCPInterface):
    """dS378 ethernet relay"""

    def __init__(self, ip, port, timeout=10):
        super().__init__(ip, port, timeout)

    def __del__(self):
        self._com.close()

    def _send(self, msg):
        self.send(msg)

    def _send1(self, msg_byte):
        self._send(bytearray([msg_byte]))

    def _recv(self):
        rcv = self.recv(LEN_BUFF)
        return rcv

    def _recv1(self):
        return self._recv()[0]

    def _send_recv(self, msg, length):
        # Check the length of the received message
        # to drop invalid response when reconnecting.
        for _ in range(N_TRIAL):
            self._send(msg)
            msg_rcv = self._recv()
            if len(msg_rcv) == length:
                return msg_rcv

        raise ConnectionError

    def get_status(self):
        """Get status of the dS378 device.

        Returns
        -------
        d_status : dict
            Status information
        """

        ret_bytes = self._send_recv(bytearray([GET_STATUS[0]]),
                                    GET_STATUS[1])

        d_status = {}
        d_status['module_id'] = ret_bytes[0]
        d_status['firm_ver'] = f'{ret_bytes[1]}.{ret_bytes[2]}'
        d_status['app_ver'] = f'{ret_bytes[3]}.{ret_bytes[4]}'
        d_status['V_sppl'] = ret_bytes[5] / 10
        d_status['T_int'] = int.from_bytes(ret_bytes[6:8], signed=True, byteorder='big') / 10

        return d_status

    def set_relay(self, relay_number, on_off, pulse_time=0):
        """Turns the relay on/off or pulses it

        Parameters
        ----------
        relay_number : int
            relay_number, 1 -- 8
        on_off : int or RelayStatus
            1: on, 0: off
        pulse_time : int, 32 bit
            See document

        Returns
        -------
        status : int
            0: ACK
            otherwise: NACK
        """
        assert 1 <= relay_number <= 8
        assert 0 <= pulse_time <= 2**32 - 1
        assert on_off in [0, 1]

        msg = bytearray([SET_RELAY[0], relay_number, on_off])
        msg += pulse_time.to_bytes(4, byteorder='big')
        self._send(msg)

        return self._recv1()

    def set_output(self, io_num, on_off):
        """Set output on/off

        Parameters
        ----------
        io_num : int, 1 -- 7
            I/O port number
        on_off : int or RelayStatus
            1: on, 0: off
        """
        assert 1 <= io_num <= 7
        assert on_off in [0, 1]

        msg = bytearray([SET_OUTPUT[0], io_num, on_off])
        self._send(msg)

        return self._recv1()

    def get_relays(self):
        """Get relay states

        Returns
        -------
        d_status : list of RelayStatus
        """

        ret_bytes = self._send_recv(bytearray([GET_RELAYS[0], 1]),
                                    GET_RELAYS[1])

        d_status = [None] * 32
        for i in range(32):
            d_status[i] = RelayStatus((ret_bytes[4 - int(i / 8)] >> (i % 8)) & 1)

        return d_status

    def get_inputs(self):
        """Get input states

        Returns
        -------
        d_status : list of RelayStatus
        """

        ret_bytes = self._send_recv(bytearray([GET_INPUTS[0], 1]),
                                    GET_INPUTS[1])

        d_status = [None] * 7
        for i in range(7):
            d_status[i] = RelayStatus((ret_bytes[1] >> i) & 1)

        return d_status

    def get_analog(self):
        """Get analog input

        Returns
        -------
        values : list of int
        """

        ret_bytes = self._send_recv(bytearray([GET_ANALOG[0]]),
                                    GET_ANALOG[1])

        values = [None] * 7
        for i in range(7):
            values[i] = int.from_bytes(ret_bytes[2 * i:2 * i + 2], byteorder='big')

        return values

    def get_counters(self, counter_number):
        """Get counters

        Parameter
        ---------
        counter_number : int
            counter number. 1 -- 8

        Returns
        -------
        c_current : int
            current counter value
        c_reg : register
            capture register for the counter
        """

        ret_bytes = self._send_recv(bytearray([GET_COUNTERS[0], counter_number]),
                                    GET_COUNTERS[1])
        c_current = int.from_bytes(ret_bytes[0:4], byteorder='big')
        c_reg = int.from_bytes(ret_bytes[4:8], byteorder='big')

        return c_current, c_reg


def main():
    """Main function"""
    ds_dev = DS378(sys.argv[1], 17123)
    print(ds_dev.get_status())
    print(ds_dev.get_relays()[:8])
    ds_dev.set_relay(3, RelayStatus.off)
    print(ds_dev.get_relays()[:8])
    print(ds_dev.get_analog())
    print(ds_dev.get_inputs())


if __name__ == '__main__':
    main()



import json
import os
import sys
import traceback
from dataclasses import asdict, dataclass, field, is_dataclass
from typing import Dict, List, Optional

import numpy as np
import sodetlib as sdl
from sodetlib.det_config import DetConfig
from sodetlib.operations import (bias_steps, bias_wave, iv, uxm_relock,
                                 uxm_setup)
from twisted.internet import defer, protocol, reactor, threads

NBIASLINES = 12


def json_safe(data):
    """Convert data so it can be serialized and decoded on
    the other end.  This includes:

    - Converting numpy arrays and scalars to generic lists and
        Python basic types.
    - Converting NaN/inf to Num (0 or +/- large number)
    """
    if is_dataclass(data):
        return json_safe(asdict(data))
    if isinstance(data, dict):
        return {k: json_safe(v) for k, v in data.items()}
    if isinstance(data, (list, tuple)):
        return [json_safe(x) for x in data]
    if hasattr(data, 'dtype'):
        # numpy arrays and scalars.
        return json_safe(data.tolist())
    if isinstance(data, (str, int, bool)):
        return data
    if isinstance(data, float):
        return np.nan_to_num(data)
    if data is None:
        return None

    print(f"Unsure how to encode data of type {type(data)}...")
    print(f"data: {data}")
    # This could still be something weird but json.dumps will
    # probably reject it!
    return data


@dataclass
class QuantileData:
    name: str
    quantiles: List[float]
    values: List[float]
    total: int

    def to_block_data(self):
        "Format quantile data for OCS HK block"
        labels = [f'{self.name}_q{int(q)}' for q in self.quantiles]
        block_data = dict(zip(labels, self.values))
        block_data[f'{self.name}_total'] = self.total
        return block_data


def compute_quantiles(name: str, arr: np.ndarray, quantiles: List[float]):
    """
    Computes QuantileData object for a given array

    """
    quantiles = [float(q) for q in quantiles]
    if np.isnan(arr).all():
        return QuantileData(name, quantiles, [0 for _ in quantiles], 0)
    qs = [float(np.nan_to_num(np.nanquantile(arr, q / 100))) for q in quantiles]
    total = int(np.sum(~np.isnan(arr)))
    return QuantileData(name, quantiles, qs, total)


def encode_dataclass(obj):
    """
    Encodes a data-class into json, replacing any json-unsafe types with
    reasonable alternatives.
    """
    data = json_safe(obj)
    return json.dumps(data).encode()


def get_smurf_control():
    """
    Get the SMuRF control object and sodetlib configuration object for the
    current slot
    """
    slot = os.environ['SLOT']
    cfg = DetConfig()
    cfg.load_config_files(slot=slot)
    S = cfg.get_smurf_control()
    S.load_tune(cfg.dev.exp['tunefile'])
    return S, cfg


def take_noise(duration, kwargs=None):
    """Runs take_noise in sodetlib"""
    if kwargs is None:
        kwargs = {}
    S, cfg = get_smurf_control()
    _, res = sdl.noise.take_noise(S, cfg, duration, **kwargs)
    wls = res['noise_pars'][:, 0]
    qs = [10, 25, 50, 75, 90]
    return {
        'quantiles': {
            'white_noise_level': compute_quantiles('white_noise_level', wls, qs),
        }
    }


def take_bgmap(kwargs=None):
    if kwargs is None:
        kwargs = {}
    S, cfg = get_smurf_control()
    bsa = bias_steps.take_bgmap(S, cfg, **kwargs)

    nchans_per_bg = [0 for _ in range(NBIASLINES + 1)]
    for bg in range(NBIASLINES):
        nchans_per_bg[bg] = int(np.sum(bsa.bgmap == bg))
    nchans_per_bg[-1] = int(np.sum(bsa.bgmap == -1))
    return {
        'nchans_per_bg': nchans_per_bg,
        'filepath': bsa.filepath,
    }


def take_iv(iv_kwargs=None):
    """Runs and analyzes an IV curve"""
    S, cfg = get_smurf_control()
    if iv_kwargs is None:
        iv_kwargs = {}
    iva = iv.take_iv(S, cfg, **iv_kwargs)
    quantiles = [10, 25, 50, 75, 90]
    return {
        'filepath': iva.filepath,
        'quantiles': {
            'Rn': compute_quantiles('Rn', iva.R_n, quantiles),
            'p_sat': compute_quantiles('p_sat', iva.p_sat, quantiles)
        }
    }


def run_uxm_setup(bands=None, kwargs=None):
    """Runs through the UXM setup procedure"""
    if kwargs is None:
        kwargs = {}
    S, cfg = get_smurf_control()
    uxm_setup.uxm_setup(S, cfg, bands=bands, **kwargs)
    return None


def run_uxm_relock(bands=None, kwargs=None):
    """Runs through the UXM relock procedure"""
    if kwargs is None:
        kwargs = {}
    S, cfg = get_smurf_control()
    _, summary = uxm_relock.uxm_relock(S, cfg, bands=bands, **kwargs)
    tr = summary['tracking_setup_results']
    data = {
        'filepath': tr.filepath,
        'all_det_num': [len(tr.is_good[tr.bands == iband]) for iband in range(0, 8)],
        'good_det_num': [len(np.where(tr.is_good[tr.bands == iband])[0]) for iband in range(0, 8)],
    }
    return data


def take_bias_steps(kwargs=None, rfrac_range=(0.2, 0.9)):
    """Takes bias steps and computes quantiles for various parameters"""
    if kwargs is None:
        kwargs = {}

    S, cfg = get_smurf_control()
    bsa = bias_steps.take_bias_steps(S, cfg, **kwargs)

    biased = np.logical_and.reduce([
        rfrac_range[0] < bsa.Rfrac,
        rfrac_range[1] > bsa.Rfrac
    ])
    qs = [10, 25, 50, 75, 90]
    data = {
        'filepath': bsa.filepath,
        'biased_total': int(np.sum(biased)),
        'biased_per_bg': [
            int(np.sum(biased[bsa.bgmap == bg])) for bg in range(12)
        ],
        'quantiles': {
            'Rfrac': compute_quantiles('Rfrac', bsa.Rfrac, qs),
            'Si': compute_quantiles('Si', bsa.Si, qs),
            'Rtes': compute_quantiles('Rtes', bsa.R0, qs),
        }
    }
    return data


def take_bias_waves(kwargs=None, rfrac_range=(0.2, 0.9)):
    if kwargs is None:
        kwargs = {}
    S, cfg = get_smurf_control()
    bwa = bias_wave.take_bias_waves(
        S, cfg, **kwargs
    )
    biased = np.logical_and.reduce([
        rfrac_range[0] < bwa.Rfrac,
        rfrac_range[1] > bwa.Rfrac
    ])
    qs = [10, 25, 50, 75, 90]
    data = {
        'filepath': bwa.filepath,
        'biased_total': int(np.sum(biased)),
        'biased_per_bg': [
            int(np.sum(biased[bwa.bgmap == bg])) for bg in range(12)
        ],
        'quantiles': {
            'Rfrac': compute_quantiles('Rfrac', bwa.Rfrac, qs),
            'Si': compute_quantiles('Si', bwa.Si, qs),
            'Rtes': compute_quantiles('Rtes', bwa.R0, qs),
        }
    }
    return data


def test():
    """Function for testing subprocess operation"""
    print("Inside test func")
    return {'test': 10}


runnable_funcs = [
    take_noise, take_iv, run_uxm_setup, run_uxm_relock, take_bias_steps,
    take_bias_waves, test, take_bgmap
]
func_map = {f.__name__: f for f in runnable_funcs}


@dataclass
class RunCfg:
    func_name: str
    "name of function to run"

    args: List = field(default_factory=list)
    "Args to pass to function"

    kwargs: Dict = field(default_factory=dict)
    "Kwargs to pass to function"

    slot: Optional[int] = None
    "Slot to setup smurf control for. If None, uses the slot from the environment variable 'SLOT'"

    run_in_main_process: bool = False
    "If true, run the function in the main process instead of a subprocess."

    def __post_init__(self):
        if self.func_name not in func_map:
            raise ValueError(f"Invalid function name: {self.func_name}")


@dataclass
class RunResult:
    success: bool
    "True if function returned succesfully"

    return_val: any = None
    "Return value from function"

    traceback: Optional[str] = None
    "Traceback if function raised an exception"


class FuncProtocol(protocol.ProcessProtocol):
    """
    Process protocol for running one of the above functions in a subprocess.
    After the subprocess is started, the RunCfg object is encoded and sent to
    the subprocess through stdin. Once the function is finished, the result is
    encoded and passed back through FD 3.
    """

    def __init__(self, cfg: RunCfg):
        self.cfg = cfg
        self.result = None

    def connectionMade(self):
        data = encode_dataclass(self.cfg)
        self.transport.write(data)
        self.transport.closeStdin()

    def childDataReceived(self, childFD, data):
        if childFD in [1, 2]:  # stdout or stderr:
            print(data.decode())

        if childFD == 3:
            self.result = json.loads(data.decode())

    def processExited(self, status):
        if self.result is None:
            raise RuntimeError("No result received from child process")

        self.deferred.callback(self.result)


@defer.inlineCallbacks
def _run_smurf_func_reactor(cfg: RunCfg) -> RunResult:
    """Helper function for run_smurf_func, that can assume reactor context"""
    prot = FuncProtocol(cfg)
    prot.deferred = defer.Deferred()
    childFDs = {0: 'w', 1: 'r', 2: 'r', 3: 'r'}  # Regular FDs plus 3 for sending results back
    env = os.environ.copy()
    if cfg.slot is not None:
        env['SLOT'] = str(cfg.slot)
    reactor.spawnProcess(
        prot, sys.executable, [sys.executable, '-u', __file__], childFDs=childFDs,
        env=env
    )
    result = yield prot.deferred
    return RunResult(**result)


def run_smurf_func(cfg: RunCfg) -> RunResult:
    """
    This function takes a RunCfg object, and runs the specified function in a
    subprocess. The result is returned as a RunResult object. This function
    must be run in a worker thread.

    Args
    -----
    cfg: RunCfg
        Configuration object to specify the function to run, and the arguments.
    """
    if cfg.run_in_main_process:
        try:
            result = RunResult(
                success=True,
                return_val=func_map[cfg.func_name](*cfg.args, **cfg.kwargs)
            )
        except Exception:
            exc = traceback.format_exc()
            print(f"Exception raised in smurf_func:\n{exc}")
            result = RunResult(
                success=False,
                traceback=exc,
            )
        return result

    return threads.blockingCallFromThread(
        reactor, _run_smurf_func_reactor, cfg)


def subprocess_main():
    """
    Starting point for subprocesses. Reads configuration info from stdin,
    runs the specified function, and returns the encoded result object through
    FD 3.
    """
    data = sys.stdin.read()
    cfg = RunCfg(**json.loads(data))
    print(f'Starting subprocess function call:\n {cfg}')
    try:
        return_val = func_map[cfg.func_name](*cfg.args, **cfg.kwargs)
        result = RunResult(success=True, return_val=return_val)
        return_data = encode_dataclass(result)
    except Exception:
        exc = traceback.format_exc()
        print(f"Exception raised in subprocess:\n{exc}")
        result = RunResult(
            success=False,
            traceback=exc,
        )
        return_data = encode_dataclass(result)
    os.write(3, return_data)


if __name__ == '__main__':
    subprocess_main()
